"""
The stop service is used to retrieve information about stops.

In terms of information retrieval, the get_in_system_by_id method is
probably the most important in Transiter. It returns a significant amount
of data, and this is reflected in the amount of code executed.
"""
import collections
import enum
import time
import typing

from transiter import exceptions, models
from transiter.data import dbconnection
from transiter.data.dams import stopdam, tripdam, systemdam
from transiter.services import views
from transiter.services.servicemap import servicemapmanager
from transiter.services.servicemap.graphutils import datastructures


@dbconnection.unit_of_work
def list_all_in_system(system_id) -> typing.List[views.Stop]:
    """
    Get information on all stops in a specific system.
    """
    system = systemdam.get_by_id(system_id, only_return_active=True)
    if system is None:
        raise exceptions.IdNotFoundError(models.System, system_id=system_id)

    return list(map(views.Stop.from_model, stopdam.list_all_in_system(system_id)))


@dbconnection.unit_of_work
def get_in_system_by_id(
    system_id,
    stop_id,
    return_links=True,
    return_only_stations=True,
    earliest_time=None,
    latest_time=None,
    minimum_number_of_trips=None,
    include_all_trips_within=None,
    exclude_trips_before=None,
):
    """
    Get information about a specific stop.
    """
    stop = stopdam.get_in_system_by_id(system_id, stop_id)
    if stop is None:
        raise exceptions.IdNotFoundError(
            models.Stop, system_id=system_id, stop_id=stop_id
        )

    stop_tree = _StopTree(stop, stopdam.list_all_stops_in_stop_tree(stop.pk))

    # The descendant stops are used as the source of trip stop times
    descendant_stop_pks = list(stop.pk for stop in stop_tree.descendents())
    direction_name_matcher = _DirectionNameMatcher(
        stopdam.list_direction_rules_for_stops(descendant_stop_pks)
    )
    trip_stop_times = stopdam.list_stop_time_updates_at_stops(
        descendant_stop_pks, earliest_time=earliest_time, latest_time=latest_time,
    )
    trip_pk_to_last_stop = tripdam.get_trip_pk_to_last_stop_map(
        trip_stop_time.trip.pk for trip_stop_time in trip_stop_times
    )

    # On the other hand, the stop tree graph that is returned consists of all
    # stations in the stop's tree
    stop_pk_to_service_maps_response = servicemapmanager.build_stop_pk_to_service_maps_response(
        stop.pk for stop in stop_tree.all_stations()
    )

    # Using the data retrieved, we then build the response
    response = views.StopLarge.from_model(stop)
    stop_tree_base: views.Stop = _build_stop_tree_response(
        stop_tree, stop_pk_to_service_maps_response, return_only_stations
    )
    response.parent_stop = stop_tree_base.parent_stop
    response.child_stops = stop_tree_base.child_stops
    response.service_maps = stop_tree_base.service_maps
    response.directions = list(direction_name_matcher.all_names())

    stop_time_filter = _TripStopTimeFilter(
        inclusion_interval_start=exclude_trips_before,
        inclusion_interval_end=include_all_trips_within,
        min_trips_per_direction=minimum_number_of_trips,
    )
    for trip_stop_time in trip_stop_times:
        direction = direction_name_matcher.match(trip_stop_time)
        if stop_time_filter.remove(trip_stop_time, direction):
            continue
        response.stop_times.append(
            _build_trip_stop_time_response(
                trip_stop_time, direction, trip_pk_to_last_stop
            )
        )
    return response


class _TripStopTimeFilter:
    def __init__(
        self, inclusion_interval_start, inclusion_interval_end, min_trips_per_direction
    ):
        self._direction_to_num_trips_so_far = {}
        self._min_trips_per_direction = min_trips_per_direction
        self._inclusion_interval_start = inclusion_interval_start
        self._inclusion_interval_end = inclusion_interval_end
        self._current_time = time.time()

    def remove(self, trip_stop_time: models.TripStopTime, direction):
        result = self._remove_helper(trip_stop_time, direction)
        if not result:
            self._direction_to_num_trips_so_far[direction] = (
                self._direction_to_num_trips_so_far.get(direction, 0) + 1
            )
        return result

    def _remove_helper(self, trip_stop_time: models.TripStopTime, direction):
        trip_time = trip_stop_time.get_time().timestamp()
        # If the trip is before the inclusion interval, remove.
        if self._inclusion_interval_start is not None and (
            trip_time <= self._current_time - float(self._inclusion_interval_start) * 60
        ):
            return True
        # If the trip is within the inclusion interval, include.
        if self._inclusion_interval_end is None or (
            trip_time <= self._current_time + float(self._inclusion_interval_end) * 60
        ):
            return False
        # If an extra trip is needed for this direction, include.
        if self._min_trips_per_direction is not None and self._direction_to_num_trips_so_far.get(
            direction, 0
        ) < int(
            self._min_trips_per_direction
        ):
            return False
        return True


def _build_trip_stop_time_response(
    trip_stop_time, direction_name, trip_pk_to_last_stop
):
    """
    Build the response for a specific trip stop time.
    """
    trip = trip_stop_time.trip
    last_stop = trip_pk_to_last_stop[trip.pk]
    result = views.TripStopTime.from_model(trip_stop_time)
    result.direction = direction_name
    result.trip = views.Trip.from_model(trip)
    result.trip.route = views.Route.from_model(trip.route)
    result.trip.last_stop = views.Stop.from_model(last_stop)
    return result


class _StopTree:
    """
    An class to represent the stop tree and desired operations on it.

    The main motivation of this data structure is to get around unavoidable lazy loading
    of tree data structures in SQLAlchemy (in fact, ORMs in general). Lazy loading means
    traversing the stops tree by using models.Stop.parent and models.Stop.children is
    inefficient: N SQL queries will be emitted where N is the number of stops in the
    tree.

    This data structure under the hood uses an adjacency list method of traversing the
    tree. The adjacency lists are constructed when the object is initialized using the
    collection of all stops in the tree. In Postgres, at least, the collection of all
    stops can be retrieved in one recursive SQL query.
    """

    class TraversalMode(enum.Enum):
        """
        Ways in which the tree can be traversed starting from the base node.
        """

        DESCENDANTS = 0
        ALL = 1

    def __init__(self, base: models.Stop, stops: typing.Iterable[models.Stop]):
        """
        Initialize a new StopTrue.

        :param base: the base of the tree. This is not necessarily the root.
        :param stops: an iterable of all stops in the tree. If stops are missing from
                      this iterable, and error will be raised.
        """
        self._base = base
        self._stop_pk_to_stop = {stop.pk: stop for stop in stops}
        self._stop_pk_to_parent_pk = {
            stop.pk: stop.parent_stop_pk for stop in self._stop_pk_to_stop.values()
        }
        self._stop_pk_to_child_pks = collections.defaultdict(list)
        for stop_pk, parent_pk in self._stop_pk_to_parent_pk.items():
            if parent_pk is not None:
                self._stop_pk_to_child_pks[parent_pk].append(stop_pk)

    def descendents(self) -> typing.Iterator[models.Stop]:
        """
        Get all descendents of the base stop, including the stop itself.
        """
        yield from self._dfs_traverse(_StopTree.TraversalMode.DESCENDANTS, False)

    def all_stations(self) -> typing.Iterator[models.Stop]:
        """
        Get all stations in the tree.

        The base stop is always returned irrespective of whether it is a station.
        """
        yield from self._dfs_traverse(_StopTree.TraversalMode.ALL, True)

    _T = typing.TypeVar("_T")

    def apply_function(
        self,
        function: typing.Callable[
            [models.Stop, typing.Optional[_T], typing.List[_T]], _T
        ],
        only_stations: bool,
    ) -> _T:
        """
        The method applies a function to each node in the traversal. It uses a
        depth-first search, so when evaluating the function at a given node
        the results of evaluating the function at the nodes the next level
        deep are available. The return of this function is the value of the node
        function at the base.

        The function signature is described above. The arguments passed to it are:

        - stop: the current node being evaluated
        - parent_result: the result of the function applied to the parent.
                         If the parent hasn't been visited in the traversal yet, this
                         will be None.
        - children_results: a list of results for the function applied to the children
                            of the current stop. Given the DFS nature of the traversal,
                            at most one child will have not been visited when the parent
                            is visited. The result of this child will, of course, be
                            missing from this list.
        """
        stop_pk_to_response = {}
        for stop in self._dfs_traverse(
            _StopTree.TraversalMode.ALL, only_stations=only_stations
        ):
            parent_response = stop_pk_to_response.get(stop.parent_stop_pk)
            children_responses = [
                stop_pk_to_response[child_pk]
                for child_pk in self._stop_pk_to_child_pks[stop.pk]
                if child_pk in stop_pk_to_response
            ]
            stop_pk_to_response[stop.pk] = function(
                stop, parent_response, children_responses
            )
        return stop_pk_to_response[self._base.pk]

    def _dfs_traverse(
        self, traversal_mode, only_stations=False
    ) -> typing.Iterator[models.Stop]:

        stack = datastructures.Stack()
        stack.push((self._base.pk, False))
        visited_pks = set()
        visited_pks.add(self._base.pk)

        while len(stack) > 0:
            stop_pk, neighbors_added = stack.pop()
            if neighbors_added:
                stop = self._stop_pk_to_stop[stop_pk]
                if only_stations and not stop.is_station and stop.pk != self._base.pk:
                    continue
                yield stop
                continue

            stack.push((stop_pk, True))
            neighbors = list(self._stop_pk_to_child_pks[stop_pk])
            neighbors.sort(reverse=True)  # Give the traversal a deterministic result
            parent_pk = self._stop_pk_to_parent_pk[stop_pk]
            if parent_pk is not None and (
                stop_pk != self._base.pk
                or traversal_mode is _StopTree.TraversalMode.ALL
            ):
                neighbors.append(parent_pk)
            for neighbor in neighbors:
                if neighbor in visited_pks:
                    continue
                visited_pks.add(neighbor)
                stack.push((neighbor, False))


def _build_stop_tree_response(
    stop_tree, stop_pk_to_service_maps_response, return_only_stations
):
    """
    Build the stop tree response.

    The response consists of a nested dictionary representing the stop tree
    with the given base as the starting point of the representation. A given
    stop in the dictionary can contain 'parent_stop' and 'child_stops'
    keys which point to relevant related stops in the tree.

    Each stop appearing also contains its short representation, its service
    map representation and, optionally, a link to the stop.
    """

    def node_function(stop, parent_return, children_return):
        response = views.Stop.from_model(stop)
        response.service_maps = stop_pk_to_service_maps_response[stop.pk]
        if parent_return is not None:
            response.parent_stop = parent_return
        if stop.parent_stop_pk is None:
            response.parent_stop = None
        if children_return is not None:
            response.child_stops = children_return
        return response

    return stop_tree.apply_function(node_function, only_stations=return_only_stations)


class _DirectionNameMatcher:
    """
    Object to find the direction name associated to a particular trip at
    a particular stop.
    """

    def __init__(self, rules):
        """
        Initialize a new matcher.

        :param rules: the rules to be used in the matcher.
        :type rules: iterable of DirectionRule models.
        """
        self._rules = sorted(rules, key=lambda rule: rule.priority)
        self._cache = {}

    def all_names(self):
        """
        Get all of the direction names in the matcher.

        :return: list of names
        :rtype: list of strings
        """
        print(self._rules)
        return {rule.name for rule in self._rules}

    def match(self, trip_stop_time):
        """
        Find the direction name associate to the TripStopTime by matching
        the appropriate rule.

        :param trip_stop_time: the TripStopTime
        :return: the direction name
        """
        cache_key = (
            trip_stop_time.stop_pk,
            trip_stop_time.trip.direction_id,
            trip_stop_time.track,
        )
        if cache_key not in self._cache:
            self._cache[cache_key] = None
            for rule in self._rules:
                if rule.stop_pk != cache_key[0]:
                    continue
                if rule.direction_id is not None and rule.direction_id != cache_key[1]:
                    continue
                if rule.track is not None and rule.track != cache_key[2]:
                    continue
                self._cache[cache_key] = rule.name
                break

        return self._cache[cache_key]



import dataclasses
import datetime
import enum
import typing

from transiter import models


class NULL:
    pass


class View:
    def to_dict(self):
        return {
            key: getattr(self, key)
            for key in self.__dataclass_fields__.keys()
            if key[0] != "_" and getattr(self, key) is not NULL
        }


@dataclasses.dataclass
class System(View):
    id: str
    status: models.System.SystemStatus
    name: str

    @classmethod
    def from_model(cls, system: models.System):
        return cls(id=system.id, status=system.status, name=system.name)


@dataclasses.dataclass
class _EntitiesInSystem(View):
    count: int
    _system_id: str

    @classmethod
    def from_model(cls, system: models.System, count):
        return cls(count=count, _system_id=system.id)


class RoutesInSystem(_EntitiesInSystem):
    pass


class StopsInSystem(_EntitiesInSystem):
    pass


class FeedsInSystem(_EntitiesInSystem):
    pass


@dataclasses.dataclass
class SystemLarge(System):
    stops: StopsInSystem = NULL
    routes: RoutesInSystem = NULL
    feeds: FeedsInSystem = NULL
    agency_alerts: list = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class SystemUpdate(View):
    id: str
    status: models.SystemUpdate.Status
    status_message: str
    scheduled_at: datetime.datetime
    completed_at: datetime.datetime
    system: System = NULL

    @classmethod
    def from_model(cls, update: models.SystemUpdate):
        return cls(
            id=str(update.pk),
            status=update.status,
            status_message=update.status_message,
            scheduled_at=update.scheduled_at,
            completed_at=update.completed_at,
        )


@dataclasses.dataclass
class Route(View):
    class Status(enum.Enum):
        NO_SERVICE = 0
        GOOD_SERVICE = 1
        PLANNED_SERVICE_CHANGE = 2
        UNPLANNED_SERVICE_CHANGE = 3
        DELAYS = 4

    id: str
    color: str
    _system_id: str
    status: Status = NULL

    @classmethod
    def from_model(cls, route: models.Route):
        return cls(id=route.id, color=route.color, _system_id=route.system.id)


@dataclasses.dataclass
class RouteLarge(View):

    id: str
    color: str
    short_name: str
    long_name: str
    description: str
    url: str
    type: models.Route.Type
    _system_id: str
    status: Route.Status
    periodicity: float
    alerts: list
    service_maps: list

    @classmethod
    def from_model(cls, route: models.Route, status, periodicity, alerts, service_maps):
        return cls(
            id=route.id,
            color=route.color,
            short_name=route.short_name,
            long_name=route.long_name,
            description=route.description,
            url=route.url,
            type=route.type,
            _system_id=route.system.id,
            status=status,
            periodicity=periodicity,
            alerts=alerts,
            service_maps=service_maps,
        )


@dataclasses.dataclass
class Stop(View):
    id: str
    name: str
    _system_id: str
    service_maps: list = NULL
    parent_stop: typing.Optional["Stop"] = NULL
    child_stops: list = NULL

    @classmethod
    def from_model(cls, stop: models.Stop):
        return cls(id=stop.id, name=stop.name, _system_id=stop.system.id)


@dataclasses.dataclass
class StopLarge(View):
    id: str
    name: str
    longitude: float
    latitude: float
    url: str
    _system_id: str
    service_maps: list = NULL
    parent_stop: typing.Optional["Stop"] = NULL
    child_stops: list = NULL
    directions: list = NULL
    stop_times: list = dataclasses.field(default_factory=list)

    @classmethod
    def from_model(cls, stop: models.Stop):
        return cls(
            id=stop.id,
            name=stop.name,
            longitude=stop.longitude,
            latitude=stop.latitude,
            url=stop.url,
            _system_id=stop.system.id,
        )


@dataclasses.dataclass
class ServiceMapWithRoutes(View):
    group_id: str
    routes: typing.List[Route]


@dataclasses.dataclass
class ServiceMapWithStops(View):
    group_id: str
    stops: typing.List[Stop]


@dataclasses.dataclass
class Trip(View):
    id: str
    direction_id: bool
    start_time: datetime.datetime
    last_update_time: datetime.datetime
    current_status: models.Trip.TripStatus
    current_stop_sequence: int
    vehicle_id: str
    _system_id: str
    _route_id: str
    route: Route = NULL
    last_stop: Stop = NULL
    stop_times: list = NULL

    @classmethod
    def from_model(cls, trip: models.Trip):
        return cls(
            id=trip.id,
            direction_id=trip.direction_id,
            start_time=trip.start_time,
            last_update_time=trip.last_update_time,
            current_status=trip.current_status,
            current_stop_sequence=trip.current_stop_sequence,
            vehicle_id=trip.vehicle_id,
            _system_id=trip.route.system.id,
            _route_id=trip.route.id,
        )


@dataclasses.dataclass
class _TripStopTimeEvent(View):
    time: datetime.datetime
    delay: typing.Optional[int]
    uncertainty: typing.Optional[int]


@dataclasses.dataclass
class TripStopTime(View):
    arrival: _TripStopTimeEvent
    departure: _TripStopTimeEvent
    track: str
    future: bool
    stop_sequence: int
    direction: str = NULL
    stop: Stop = NULL
    trip: Trip = NULL

    @classmethod
    def from_model(cls, trip_stop_time: models.TripStopTime):
        return cls(
            arrival=_TripStopTimeEvent(
                time=trip_stop_time.arrival_time,
                delay=trip_stop_time.arrival_delay,
                uncertainty=trip_stop_time.arrival_uncertainty,
            ),
            departure=_TripStopTimeEvent(
                time=trip_stop_time.departure_time,
                delay=trip_stop_time.departure_delay,
                uncertainty=trip_stop_time.departure_uncertainty,
            ),
            track=trip_stop_time.track,
            future=trip_stop_time.future,
            stop_sequence=trip_stop_time.stop_sequence,
        )


@dataclasses.dataclass
class Feed(View):
    id: str
    auto_update_period: int
    _system_id: str
    updates: "UpdatesInFeedLink" = NULL
    system: models.System = NULL

    @classmethod
    def from_model(cls, feed: models.Feed, add_system=False):
        return cls(
            id=feed.id,
            auto_update_period=feed.auto_update_period,
            _system_id=feed.system.id,
            system=System.from_model(feed.system) if add_system else NULL,
        )


@dataclasses.dataclass
class FeedLarge(Feed):
    pass


@dataclasses.dataclass
class UpdatesInFeedLink(View):
    _feed_id: str
    _system_id: str

    @classmethod
    def from_model(cls, feed: models.Feed):
        return cls(_feed_id=feed.id, _system_id=feed.system.id)


@dataclasses.dataclass
class FeedUpdate(View):
    id: str
    type: models.FeedUpdate.Type
    status: models.FeedUpdate.Status
    result: models.FeedUpdate.Result
    result_message: str
    content_hash: str
    content_length: int
    completed_at: datetime.datetime

    @classmethod
    def from_model(cls, feed_update: models.FeedUpdate):
        return cls(
            id=str(feed_update.pk),
            type=feed_update.update_type,
            status=feed_update.status,
            result=feed_update.result,
            result_message=feed_update.result_message,
            content_hash=feed_update.content_hash,
            content_length=feed_update.content_length,
            completed_at=feed_update.completed_at,
        )


@dataclasses.dataclass
class AlertLarge(View):
    id: str
    start_time: datetime.datetime
    end_time: datetime.datetime
    creation_time: datetime.datetime
    header: str
    description: str
    url: str
    cause: models.Alert.Cause
    effect: models.Alert.Effect

    @classmethod
    def from_model(cls, alert: models.Alert):
        return cls(
            id=alert.id,
            start_time=alert.start_time,
            end_time=alert.end_time,
            creation_time=alert.creation_time,
            header=alert.header,
            description=alert.description,
            url=alert.url,
            cause=alert.cause,
            effect=alert.effect,
        )



from sqlalchemy import Column, Integer, ForeignKey, Index
from sqlalchemy.orm import relationship

from .base import Base


class ServiceMapVertex(Base):
    __tablename__ = "service_map_vertex"

    pk = Column(Integer, primary_key=True)
    stop_pk = Column(Integer, ForeignKey("stop.pk"))
    map_pk = Column(Integer, ForeignKey("service_map.pk"))

    position = Column(Integer)

    stop = relationship("Stop", cascade=None)
    map = relationship("ServiceMap", back_populates="vertices", cascade=None)

    __table_args__ = (Index("service_map_vertex_map_pk_position", map_pk, position),)



# When updating the version, also update it in the docs.
__version__ = "0.4.5"

# The following are set by Travis CI at build time.
__build_number__ = None
__build_timestamp__ = None
__build_href__ = None
__git_commit_hash__ = None

# NOTE: the following were set by Travis CI
__version__ = "0.4.5"
__build_number__ = 850
__build_timestamp__ = 1589467080
__build_href__ = "https://travis-ci.org/jamespfennell/transiter/jobs/687048115"
__git_commit_hash__ = "74e61ad1df2e2d56308128158ac399c444be9d81"



"""
This module is responsible for the database session and transaction scope.
"""
import logging
from contextlib import contextmanager

import sqlalchemy.exc
from alembic import command
from alembic.config import Config
from decorator import decorator
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import scoped_session, sessionmaker

from transiter import models, config

logger = logging.getLogger(__name__)


def create_engine():
    """
    Create a SQL Alchemy engine using config.DatabaseConfig.

    :return: the engine
    """
    connection_url = URL(
        drivername=config.DB_DRIVER,
        username=config.DB_USERNAME,
        password=config.DB_PASSWORD,
        host=config.DB_HOST,
        port=config.DB_PORT,
        database=config.DB_DATABASE,
    )
    return sqlalchemy.create_engine(
        connection_url,
        executemany_mode="values",
        executemany_values_page_size=1000,
        executemany_batch_page_size=200,
    )


engine = None
session_factory = None
Session = None


def ensure_db_connection():
    """
    Ensure that the SQL Alchemy engine and session factory have been initialized.
    """
    global engine, session_factory, Session
    if engine is not None:
        return
    engine = create_engine()
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)


class OutsideUnitOfWorkError(Exception):
    pass


class NestedUnitOfWorkError(Exception):
    pass


def get_session():
    """
    Get the current session.

    :return: the session
    :raises OutsideUnitOfWorkError: if this method is called from outside a UOW
    """
    global Session
    if Session is None or not Session.registry.has():
        raise OutsideUnitOfWorkError
    return Session()


@contextmanager
def inline_unit_of_work():
    """
    Context manager that handles beginning and ending a unit of work.
    """
    global Session
    ensure_db_connection()
    if Session.registry.has():
        raise NestedUnitOfWorkError
    session = Session()
    try:
        yield session
        session.commit()
    except BaseException:
        session.rollback()
        raise
    finally:
        Session.remove()


@decorator
def unit_of_work(func, *args, **kw):
    """
    Decorator that handles beginning and ending a unit of work.
    """
    with inline_unit_of_work():
        return func(*args, **kw)


def delete_all_tables():
    """
    Delete all Transiter tables in the database.
    """
    global engine
    ensure_db_connection()
    models.Base.metadata.drop_all(engine)
    alembic_config = _get_alembic_config()
    command.stamp(alembic_config, None)


def upgrade_database():
    command.upgrade(_get_alembic_config(), "head")


def get_current_database_revision():
    captured_text = None

    def print_stdout(text, *arg, **kwargs):
        nonlocal captured_text
        captured_text = text

    alembic_config = _get_alembic_config()
    alembic_config.print_stdout = print_stdout
    command.current(alembic_config)
    return captured_text


def _get_alembic_config() -> Config:
    alembic_config = Config()
    alembic_config.set_main_option("script_location", "transiter:alembic")
    return alembic_config



from sqlalchemy import (
    Column,
    Integer,
    String,
    ForeignKey,
    Boolean,
    Numeric,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship, backref

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.Stop)
class Stop(Base):
    __tablename__ = "stop"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    system_pk = Column(Integer, ForeignKey("system.pk"), nullable=False)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)
    parent_stop_pk = Column(Integer, ForeignKey("stop.pk"), index=True)

    name = Column(String)
    longitude = Column(Numeric(precision=9, scale=6))
    latitude = Column(Numeric(precision=9, scale=6))
    url = Column(String)
    is_station = Column(Boolean)

    system = relationship("System", back_populates="stops")
    source = relationship("FeedUpdate", cascade="none")
    # NOTE: this relationship is a little tricky and this definition follows
    # https://docs.sqlalchemy.org/en/latest/_modules/examples/adjacency_list/adjacency_list.html
    child_stops = relationship(
        "Stop",
        cascade="none",
        backref=backref("parent_stop", remote_side=pk, cascade="none"),
    )
    direction_rules = relationship(
        "DirectionRule",
        back_populates="stop",
        cascade="all, delete-orphan",
        order_by="DirectionRule.priority",
    )
    trip_times = relationship(
        "TripStopTime", back_populates="stop", cascade="delete, delete-orphan"
    )
    service_pattern_vertices = relationship(
        "ServiceMapVertex", back_populates="stop", cascade="delete, delete-orphan"
    )
    scheduled_trip_times = relationship(
        "ScheduledTripStopTime", back_populates="stop", cascade="delete, delete-orphan"
    )

    __table_args__ = (UniqueConstraint("system_pk", "id"),)

    @staticmethod
    def from_parsed_stop(stop: parse.Stop) -> "Stop":
        return Stop(
            id=stop.id,
            name=stop.name,
            latitude=stop.latitude,
            longitude=stop.longitude,
            url=stop.url,
            is_station=stop.is_station,
        )



from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.ScheduledService)
class ScheduledService(Base):
    __tablename__ = "scheduled_service"

    pk = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    system_pk = Column(Integer, ForeignKey("system.pk"), index=True)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    monday = Column(Boolean)
    tuesday = Column(Boolean)
    wednesday = Column(Boolean)
    thursday = Column(Boolean)
    friday = Column(Boolean)
    saturday = Column(Boolean)
    sunday = Column(Boolean)

    source = relationship("FeedUpdate", cascade="none")
    system = relationship("System", back_populates="scheduled_services")
    trips = relationship(
        "ScheduledTrip", back_populates="service", cascade="delete, delete-orphan"
    )

    @staticmethod
    def from_parsed_service(service: parse.ScheduledService) -> "ScheduledService":
        return ScheduledService(
            id=service.id,
            monday=service.monday,
            tuesday=service.tuesday,
            wednesday=service.wednesday,
            thursday=service.thursday,
            friday=service.friday,
            saturday=service.saturday,
            sunday=service.sunday,
        )



import datetime
import enum

from sqlalchemy import (
    Column,
    TIMESTAMP,
    Index,
    Float,
    Enum,
    Integer,
    String,
    ForeignKey,
)
from sqlalchemy.orm import relationship

from .base import Base


class FeedUpdate(Base):
    __tablename__ = "feed_update"

    pk = Column(Integer, primary_key=True)
    feed_pk = Column(Integer, ForeignKey("feed.pk"))

    class Type(enum.Enum):
        REGULAR = 1
        FLUSH = 2

    class Status(enum.Enum):
        SCHEDULED = 1
        IN_PROGRESS = 2
        SUCCESS = 3
        FAILURE = 4

    class Result(enum.Enum):
        UPDATED = 1
        NOT_NEEDED = 2
        PARSE_ERROR = 3
        DOWNLOAD_ERROR = 4
        INVALID_PARSER = 5
        EMPTY_FEED = 6
        SYNC_ERROR = 7
        UNEXPECTED_ERROR = 8

    update_type = Column(
        Enum(Type, name="update_type", native_enum=False),
        nullable=False,
        default=Type.REGULAR,
    )
    status = Column(Enum(Status, native_enum=False))
    result = Column(Enum(Result, native_enum=False))
    result_message = Column(String)
    content_hash = Column(String)
    content_length = Column(Integer)
    content_created_at = Column(TIMESTAMP(timezone=True))
    download_duration = Column(Float)
    total_duration = Column(Float)
    scheduled_at = Column(TIMESTAMP(timezone=True), default=datetime.datetime.utcnow)
    completed_at = Column(TIMESTAMP(timezone=True))
    num_parsed_entities = Column(Integer)
    num_added_entities = Column(Integer)
    num_updated_entities = Column(Integer)
    num_deleted_entities = Column(Integer)

    feed = relationship("Feed", back_populates="updates")

    __table_args__ = (
        Index(
            "feed_update_success_pk_completed_at_idx",
            feed_pk,
            completed_at,
            postgresql_where=(status == Status.SUCCESS),
        ),
        Index("feed_update_feed_pk_feed_update_pk_idx", feed_pk, pk),
    )



# This module provides a registry for updatable entities.


def updatable_from(feed_entity):
    def decorator(cls):
        _db_entity_to_feed_entity[cls] = feed_entity
        return cls

    return decorator


def list_updatable_entities():
    return list(_db_entity_to_feed_entity.keys())


def list_feed_entities():
    return list(_db_entity_to_feed_entity.values())


def get_feed_entity(db_entity):
    return _db_entity_to_feed_entity[db_entity]


_db_entity_to_feed_entity = {}



from typing import Dict, Iterable

import sqlalchemy.sql.expression as sql
from sqlalchemy.orm import selectinload

from transiter import models
from transiter.data import dbconnection


def list_groups_and_maps_for_stops_in_route(route_pk):
    """
    This function is used to get the service maps for a route.

    It returns a list of tuples (service map group, service map) for each
    service map group having use_for_stops_in_route equal True.

    :param route_pk: the route's PK
    :return: the list described above
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.ServiceMapGroup, models.ServiceMap)
        .join(models.System, models.System.pk == models.ServiceMapGroup.system_pk)
        .join(models.Route, models.Route.system_pk == models.System.pk)
        .outerjoin(
            models.ServiceMap,
            sql.and_(
                models.ServiceMap.route_pk == models.Route.pk,
                models.ServiceMap.group_pk == models.ServiceMapGroup.pk,
            ),
        )
        .filter(models.ServiceMapGroup.use_for_stops_in_route)
        .filter(models.Route.pk == route_pk)
        .options(selectinload(models.ServiceMap.vertices))
        .options(selectinload(models.ServiceMap.vertices, models.ServiceMapVertex.stop))
    )
    return [(group, map_) for (group, map_) in query]


def get_stop_pk_to_group_id_to_routes_map(
    stop_pks,
) -> Dict[int, Dict[str, Iterable[models.Route]]]:
    """
    This function is used to get service map information for stops; namely,
    which routes call at the stop based on the service maps.

    Get a map whose key is a stop's PK and whose the value is another map.
    This second map has a key for every service map group having
    use_for_routes_at_stop equal to True. The value of this map is the list of
    routes that contain the stop in the relevant service map.

    :param stop_pks: stop PKs to build the map for
    :return: the monster map described above
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.Stop.pk, models.ServiceMapGroup.id, models.Route)
        .join(models.System, models.System.pk == models.Stop.system_pk)
        .join(
            models.ServiceMapGroup,
            sql.and_(
                models.ServiceMapGroup.system_pk == models.System.pk,
                models.ServiceMapGroup.use_for_routes_at_stop,
            ),
        )
        .outerjoin(
            models.ServiceMap,
            sql.and_(
                models.ServiceMap.group_pk == models.ServiceMapGroup.pk,
                models.ServiceMap.pk.in_(
                    session.query(models.ServiceMapVertex.map_pk).filter(
                        models.ServiceMapVertex.stop_pk == models.Stop.pk
                    )
                ),
            ),
        )
        .outerjoin(models.Route, models.Route.pk == models.ServiceMap.route_pk)
        .filter(models.Stop.pk.in_(stop_pks))
    )
    response = {stop_pk: {} for stop_pk in stop_pks}
    for stop_pk, group_id, route in query:
        if group_id not in response[stop_pk]:
            response[stop_pk][group_id] = []
        if route is not None:
            response[stop_pk][group_id].append(route)
    return response



"""
This class contains algorithms for topologically sorting graphs.
"""

from transiter.services.servicemap.graphutils import (
    datastructures,
    operations,
    traversals,
)


class ImpossibleToTopologicallySortGraph(Exception):
    """
    Exception thrown if the inputted directed graph contains a cycle.
    """


def basic_sort(graph: datastructures.AbstractGraph) -> datastructures.OrderedGraph:
    """
    Topologically sort a directed graph.

    The algorithm here is the basic topological sort algorithm, except the
    sorting has the following property. If there are vertices A, B, C, such
    (A,B) and (B,C) are edges, and there are no other edges out of A and B
    and in to B and C, then A, B and C will be consecutive in the sorting.

    Raises ImpossibleToTopologicallySortGraph: if the graph cannot be sorted;
    i.e., it contains a cycle.
    """
    sorted_nodes = []
    admissible_next_nodes = datastructures.Stack(graph.sources())
    node_to_in_node_count = {node: len(node.in_nodes) for node in graph.nodes()}

    while len(admissible_next_nodes) > 0:
        node = admissible_next_nodes.pop()
        sorted_nodes.append(node)
        for candidate_next_node in node.out_nodes:
            node_to_in_node_count[candidate_next_node] -= 1
            if node_to_in_node_count[candidate_next_node] == 0:
                admissible_next_nodes.push(candidate_next_node)

    if len(graph) != len(sorted_nodes):
        raise ImpossibleToTopologicallySortGraph

    return datastructures.OrderedGraph(graph, sorted_nodes)


def optimal_sort_for_trees(graph: datastructures.Tree) -> datastructures.OrderedTree:
    graph_reversed = False
    if graph.is_in_tree():
        graph.reverse()
        graph_reversed = True
    node_to_descendents_count = {}
    for node in traversals.post_order_dfs_traversal(graph.root):
        node_to_descendents_count[node] = 1 + sum(
            node_to_descendents_count[child_node] for child_node in node.out_nodes
        )
    ordered_graph = datastructures.OrderedTree(
        graph,
        traversals.pre_order_dfs_traversal(
            graph.root, sorting_key=node_to_descendents_count.get
        ),
    )
    if graph_reversed:
        ordered_graph.reverse()
    return ordered_graph


def tgt_sort(graph: datastructures.AbstractGraph) -> datastructures.OrderedGraph:
    tree_1, inner_graph, tree_2 = operations.calculate_tgt_decomposition(
        operations.calculate_transitive_reduction(graph.immutable())
    )

    labels = []
    for sub_graph, sorting_method in (
        (tree_1, optimal_sort_for_trees),
        (inner_graph, basic_sort),
        (tree_2, optimal_sort_for_trees),
    ):
        if sub_graph is None:
            continue
        new_labels = [node.label for node in sorting_method(sub_graph).nodes()]
        if len(labels) > 0:
            assert labels[-1] == new_labels[0]
            labels += new_labels[1:]
        else:
            labels = new_labels
    return datastructures.OrderedGraph(
        graph, [graph.get_node(label) for label in labels]
    )



import logging
import os

from celery import Celery, signals

host = os.environ.get("TRANSITER_RABBITMQ_HOST", "127.0.0.1")

app = Celery("transiter", broker="amqp://{}".format(host))


@signals.after_setup_logger.connect
def on_after_setup_logger(**kwargs):
    # These configurations disable the task received at and completed at msgs
    logger = logging.getLogger("celery.worker.strategy")
    logger.propagate = False
    logger = logging.getLogger("celery.app.trace")
    logger.propagate = False


def run(log_level="warning"):
    app.autodiscover_tasks(
        packages=["transiter.services.systemservice", "transiter.scheduler.server"],
        related_name=None,
    )
    app.start(argv=["celeryapp", "worker", "-l", log_level])



import logging

from flask import Blueprint

from transiter.data import dbconnection
from transiter.executor.celeryapp import app as celery_app
from transiter.http.httpmanager import http_endpoint, HttpMethod
from transiter.http.permissions import requires_permissions, PermissionsLevel
from transiter.scheduler import client

admin_endpoints = Blueprint(__name__, __name__)

logger = logging.getLogger(__name__)


@http_endpoint(admin_endpoints, "health")
@requires_permissions(PermissionsLevel.ADMIN_READ)
def health():
    """Generate a health response."""
    scheduler_response = client.ping()
    if scheduler_response is None:
        scheduler_up = False
        scheduler_num_tasks = None
    else:
        scheduler_up = True
        scheduler_num_tasks = len(scheduler_response)
    num_celery_workers = len(celery_app.control.ping(timeout=0.25))
    celery_up = num_celery_workers > 0
    return {
        "up": celery_up and scheduler_up,
        "services": {
            "scheduler": {"up": scheduler_up, "num_update_tasks": scheduler_num_tasks},
            "executor": {"up": celery_up, "num_workers": num_celery_workers},
        },
    }


@http_endpoint(admin_endpoints, "scheduler")
@requires_permissions(PermissionsLevel.ADMIN_READ)
def scheduler_ping():
    return client.ping()


@http_endpoint(admin_endpoints, "scheduler", method=HttpMethod.POST)
@requires_permissions(PermissionsLevel.ALL)
def scheduler_refresh_tasks():
    return client.refresh_tasks()


@http_endpoint(admin_endpoints, "upgrade", method=HttpMethod.POST)
@requires_permissions(PermissionsLevel.ALL)
def upgrade_database():
    return dbconnection.upgrade_database()



"""Make enum non-native

Revision ID: 5bc1207e7593
Revises: 055583d82811
Create Date: 2020-04-23 10:30:45.143093

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5bc1207e7593"
down_revision = "055583d82811"
branch_labels = None
depends_on = None


def upgrade():
    convert_enum_to_non_native("system", "status", "ACTIVE")
    convert_enum_to_non_native("feed", "built_in_parser")
    convert_enum_to_non_native("service_map_group", "source", "SCHEDULE")
    convert_enum_to_non_native("feed_update", "update_type", "REGULAR")
    convert_enum_to_non_native("feed_update", "status")
    convert_enum_to_non_native("alert", "cause")
    convert_enum_to_non_native("alert", "effect")
    convert_enum_to_non_native("route", "type")
    convert_enum_to_non_native("trip", "current_status")

    old_enum_names = [
        "systemstatus",
        "builtinparser",
        "servicemapsource",
        "update_type",
        "status",
        "explanation",
        "cause",
        "effect",
        "route_type",
        "tripstatus",
    ]
    for enum_name in old_enum_names:
        sa.Enum(name=enum_name).drop(op.get_bind(), checkfirst=True)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def convert_enum_to_non_native(table_name, column_name, default=None):
    temp_column_name = column_name + "_temp"

    # Move old column to temporary new name
    op.alter_column(table_name, column_name, new_column_name=temp_column_name)

    # Create new column with old columns name
    op.add_column(
        table_name, sa.Column(column_name, sa.String(), nullable=True),
    )

    # If there is a default set, populate all fields with the default
    if default is not None:
        op.get_bind().execute(f"update {table_name} set {column_name} = '{default}'")
        op.alter_column(table_name, column_name, nullable=False)

    # Migrate the data over
    op.get_bind().execute(f"update {table_name} set {column_name} = {temp_column_name}")

    # Drop old column
    op.drop_column(table_name, temp_column_name)



from typing import Optional

from sqlalchemy import func, sql
from sqlalchemy.orm import joinedload

from transiter import models
from transiter.data import dbconnection
from transiter.data.dams import genericqueries


def create(entity=None):
    """
    Create a new System.

    :param entity: An optional System to persist
    :return: the persisted System
    """
    return genericqueries.create(models.System, entity)


def delete_by_id(id_):
    """
    Delete a System from the DB whose ID is given.

    :return: True if an entity was found and deleted, false if no such
     entity exists
    """
    session = dbconnection.get_session()
    entity = get_by_id(id_)
    if entity is None:
        return False
    session.delete(entity)
    return True


def list_all():
    """
    List all Systems in the database.
    """
    return genericqueries.list_all(models.System, models.System.id)


def get_by_id(id_, only_return_active=False) -> Optional[models.System]:
    """
    Get a system by its ID.
    """
    system = genericqueries.get_by_id(models.System, id_)
    if system is None:
        return None
    if only_return_active and system.status != system.SystemStatus.ACTIVE:
        return None
    return system


def get_update_by_pk(pk) -> Optional[models.SystemUpdate]:
    return (
        dbconnection.get_session()
        .query(models.SystemUpdate)
        .filter(models.SystemUpdate.pk == pk)
        .options(joinedload(models.SystemUpdate.system))
        .one_or_none()
    )


def set_auto_update_enabled(system_id, auto_update_enabled) -> bool:
    """
    Set the auto update enabled flag for a system.

    Returns a boolean denoting whether the system exists.
    """
    session = dbconnection.get_session()
    system_exists = session.query(
        sql.exists().where(models.System.id == system_id)
    ).scalar()
    if not system_exists:
        return False
    (
        session.query(models.System)
        .filter(models.System.id == system_id)
        .update({"auto_update_enabled": auto_update_enabled})
    )
    return True


def _count_child_entity_in_system(system_id, Model):
    """
    Return the number of entities of a certain type that are children of a System.

    :param system_id: the system's ID
    :param Model: the model type: Trip, Route or Stop
    :return: the integer count
    """
    session = dbconnection.get_session()
    query = (
        session.query(func.count(Model.pk))
        .filter(Model.system_pk == models.System.pk)
        .filter(models.System.id == system_id)
    )
    return query.one()[0]


def count_stops_in_system(system_id):
    """
    Return the number of Stops in a system.

    :param system_id: the system's ID
    :return: the integer count
    """
    return _count_child_entity_in_system(system_id, models.Stop)


def count_routes_in_system(system_id):
    """
    Return the number of Routes in a system.

    :param system_id: the system's ID
    :return: the integer count
    """
    return _count_child_entity_in_system(system_id, models.Route)


def count_feeds_in_system(system_id):
    """
    Return the number of Feeds in a system.

    :param system_id: the system's ID
    :return: the integer count
    """
    return _count_child_entity_in_system(system_id, models.Feed)


def list_all_alerts_associated_to_system(system_pk):
    """
    List all of the alerts associated directly to a system.

    This does *not* return alerts that are associated other entities in the system

    :param system_id: the system's ID
    :return: list of Alerts
    """
    session = dbconnection.get_session()
    query = session.query(models.Alert).filter(models.Alert.system_pk == system_pk)
    return query.all()



from .alert import Alert, alert_route
from .base import Base
from .directionrule import DirectionRule
from .feed import Feed
from .feedupdate import FeedUpdate
from .route import Route
from .scheduledservice import ScheduledService
from .scheduledtrip import ScheduledTrip
from .scheduledtripstoptime import ScheduledTripStopTime
from .servicemap import ServiceMap
from .servicemapgroup import ServiceMapGroup
from .servicemapvertex import ServiceMapVertex
from .stop import Stop
from .system import System
from .systemupdate import SystemUpdate
from .trip import Trip
from .tripstoptime import TripStopTime
from .updatableentity import list_updatable_entities



import typing

from transiter import parse


class TripDataCleaner:
    """
    The TripDataCleaner provides a mechanism for cleaning valid and removing
    invalid Trips and TripStopTimes.

    To use the TripDataCleaner, you must provide a number of cleaning functions.
    Trip cleaning functions accept two arguments - the current FeedUpdate
    and the Trip being cleaned,
    and perform some operations to clean up the trip - for example, switching
    the direction of Trips in a given route to compensate for a known bug in a
    Transit agency's data feed. If the cleaner returns
    False, then that Trip is removed from the collection. TripStopTime cleaners
    work identically.

    After initializing the cleaner with cleaning functions, a list of Trips
    is passed into its clean method. The cleaners operate on all of the Trips
    and their contained TripStopTimes, remove entities based on cleaner function
    results, and returns the list of cleaned trips.
    """

    def __init__(
        self,
        trip_cleaners: typing.List[typing.Callable[[parse.Trip], bool]],
        stop_time_cleaners: typing.List[
            typing.Callable[[parse.Trip, parse.TripStopTime], bool]
        ],
    ):
        """
        Initialize a new TripDataCleaner

        :param trip_cleaners: list of Trip cleaning functions
        :param stop_time_cleaners: list of TripStopTime cleaning functions
        """
        self._trip_cleaners = trip_cleaners
        self._stop_time_cleaners = stop_time_cleaners

    def clean(self, trips):
        """
        Clean a collection of trips.

        :param trips: the trips to clean
        :return: the cleaned trips with bad trips removed
        """
        trips_to_keep = []
        for trip in trips:
            if not isinstance(trip, parse.Trip):
                trips_to_keep.append(trip)
                continue
            result = True
            for trip_cleaner in self._trip_cleaners:
                result = trip_cleaner(trip)
                if not result:
                    break
            if not result:
                continue

            for stop_time_update in trip.stop_times:
                for stop_time_cleaner in self._stop_time_cleaners:
                    stop_time_cleaner(trip, stop_time_update)

            trips_to_keep.append(trip)

        return trips_to_keep



"""
The feed service is used to retrieve data about feeds and perform feed update
operations.
"""

import datetime
import logging
import typing

from transiter import exceptions, models
from transiter.data import dbconnection
from transiter.data.dams import feeddam, systemdam
from transiter.services import views
from transiter.services.update import updatemanager

logger = logging.getLogger(__name__)


@dbconnection.unit_of_work
def list_all_auto_updating() -> typing.List[views.Feed]:
    """
    List all auto updating feeds.
    """
    return [
        views.Feed.from_model(feed, add_system=True)
        for feed in feeddam.list_all_auto_updating()
    ]


@dbconnection.unit_of_work
def list_all_in_system(system_id) -> typing.List[views.Feed]:
    """
    Get data on all feeds in a system.
    """
    system = systemdam.get_by_id(system_id, only_return_active=True)
    if system is None:
        raise exceptions.IdNotFoundError(models.System, system_id=system_id)
    return list(map(views.Feed.from_model, feeddam.list_all_in_system(system_id)))


@dbconnection.unit_of_work
def get_in_system_by_id(system_id, feed_id) -> views.Feed:
    """
    Get data on a specific feed in a system.
    """
    feed = feeddam.get_in_system_by_id(system_id, feed_id)
    if feed is None:
        raise exceptions.IdNotFoundError(
            models.Feed, system_id=system_id, feed_id=feed_id
        )
    response = views.FeedLarge.from_model(feed)
    response.updates = views.UpdatesInFeedLink.from_model(feed)
    return response


def create_and_execute_feed_update(
    system_id, feed_id, content=None, execute_async=False
):
    return _create_and_execute_feed_update_helper(
        system_id,
        feed_id,
        update_manager_function=updatemanager.create_feed_update,
        content=content,
        execute_async=execute_async,
    )


def create_and_execute_feed_flush(system_id, feed_id, execute_async=False):
    return _create_and_execute_feed_update_helper(
        system_id,
        feed_id,
        update_manager_function=updatemanager.create_feed_flush,
        content=None,
        execute_async=execute_async,
    )


def _create_and_execute_feed_update_helper(
    system_id, feed_id, update_manager_function, content=None, execute_async=False,
):
    """
    Create a feed update for a feed in a system.
    """
    feed_update_pk = update_manager_function(system_id, feed_id)
    if feed_update_pk is None:
        raise exceptions.IdNotFoundError(
            models.Feed, system_id=system_id, feed_id=feed_id
        )
    if execute_async:
        updatemanager.execute_feed_update_async.delay(feed_update_pk, content)
    else:
        updatemanager.execute_feed_update(feed_update_pk, content)
    return feed_update_pk


@dbconnection.unit_of_work
def list_updates_in_feed(system_id, feed_id):
    """
    List all of the updates for a feed.
    """
    feed = feeddam.get_in_system_by_id(system_id, feed_id)
    if feed is None:
        raise exceptions.IdNotFoundError(
            models.Feed, system_id=system_id, feed_id=feed_id
        )
    response = []
    for feed_update in feeddam.list_updates_in_feed(feed.pk):
        response.append(views.FeedUpdate.from_model(feed_update))
    return response


@dbconnection.unit_of_work
def get_update_in_feed_by_pk(system_id, feed_id, feed_update_pk):
    feed_update = feeddam.get_update_by_pk(feed_update_pk)
    if feed_update is None:
        raise exceptions.IdNotFoundError(
            models.Feed,
            system_id=system_id,
            feed_id=feed_id,
            feed_update_id=str(feed_update_pk),
        )
    return views.FeedUpdate.from_model(feed_update)


def trim_feed_updates():
    """
    Delete old feed updates.

    This method is designed to be called hourly by the task server.
    """

    @dbconnection.unit_of_work
    def _list_all_feed_pks():
        return feeddam.list_all_feed_pks()

    @dbconnection.unit_of_work
    def _trim_feed_updates_helper(feed_pk_, before_datetime_):
        logger.info(
            "Deleting feed updates with feed_pk={} and last updated before {}".format(
                feed_pk_, before_datetime_
            )
        )
        feeddam.trim_feed_updates(feed_pk_, before_datetime_)

    logger.info("Trimming old feed updates.")
    before_datetime = (
        datetime.datetime.now() - datetime.timedelta(minutes=60)
    ).replace(microsecond=0, second=0)

    for feed_pk in _list_all_feed_pks():
        _trim_feed_updates_helper(feed_pk, before_datetime)



from sqlalchemy import Column, TIMESTAMP, Table, Integer, String, ForeignKey, Enum
from sqlalchemy.orm import relationship

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.Alert)
class Alert(Base):
    __tablename__ = "alert"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    Cause = parse.Alert.Cause
    Effect = parse.Alert.Effect

    header = Column(String)
    description = Column(String)
    url = Column(String)

    cause = Column(Enum(Cause, native_enum=False))
    effect = Column(Enum(Effect, native_enum=False))
    priority = Column(Integer)
    start_time = Column(TIMESTAMP(timezone=True))
    end_time = Column(TIMESTAMP(timezone=True))
    creation_time = Column(TIMESTAMP(timezone=True))

    # NOTE: the following foreign key is *temporary*! It is used when, in a feed, the
    # alert's selector references an agency. Pending the introduction of a
    # models.Agency type with which we can many-to-many relate, we just relate such an
    # alert to the system it is in.
    system_pk = Column(Integer, ForeignKey("system.pk"), index=True)

    route_ids = set()
    agency_ids = set()

    source = relationship("FeedUpdate", cascade="none")
    routes = relationship(
        "Route",
        secondary="alert_route",
        back_populates="route_statuses",
        cascade="none",
    )

    @staticmethod
    def from_parsed_alert(alert: parse.Alert) -> "Alert":
        return Alert(
            id=alert.id,
            cause=alert.cause,
            effect=alert.effect,
            header=alert.header,
            description=alert.description,
            url=alert.url,
            start_time=alert.start_time,
            end_time=alert.end_time,
            priority=alert.priority,
        )


alert_route = Table(
    "alert_route",
    Base.metadata,
    Column("alert_pk", Integer, ForeignKey("alert.pk")),
    Column("route_pk", Integer, ForeignKey("route.pk"), index=True),
)



from flask import Blueprint

from transiter.http.httpmanager import (
    http_endpoint,
    link_target,
    HttpMethod,
    HttpStatus,
    is_sync_request,
)
from transiter.http.permissions import requires_permissions, PermissionsLevel
from transiter.services import feedservice, views

feed_endpoints = Blueprint(__name__, __name__)


@http_endpoint(feed_endpoints, "")
@link_target(views.FeedsInSystem, ["_system_id"])
@requires_permissions(PermissionsLevel.ADMIN_READ)
def list_all_in_system(system_id):
    """List all feeds for a specific system."""
    return feedservice.list_all_in_system(system_id)


@http_endpoint(feed_endpoints, "/<feed_id>")
@link_target(views.Feed, ["_system_id", "id"])
@requires_permissions(PermissionsLevel.ADMIN_READ)
def get_in_system_by_id(system_id, feed_id):
    """Retrieve a specific feed."""
    return feedservice.get_in_system_by_id(system_id, feed_id)


@http_endpoint(feed_endpoints, "/<feed_id>", method=HttpMethod.POST)
@requires_permissions(PermissionsLevel.ALL)
def create_feed_update(system_id, feed_id):
    feed_update_pk = feedservice.create_and_execute_feed_update(
        system_id, feed_id, execute_async=not is_sync_request()
    )
    return (
        feedservice.get_update_in_feed_by_pk(system_id, feed_id, feed_update_pk),
        HttpStatus.CREATED,
    )


@http_endpoint(feed_endpoints, "/<feed_id>/flush", method=HttpMethod.POST)
@requires_permissions(PermissionsLevel.ALL)
def create_feed_update_flush(system_id, feed_id):
    feed_update_pk = feedservice.create_and_execute_feed_flush(
        system_id, feed_id, execute_async=not is_sync_request()
    )
    return (
        feedservice.get_update_in_feed_by_pk(system_id, feed_id, feed_update_pk),
        HttpStatus.CREATED,
    )


@http_endpoint(feed_endpoints, "/<feed_id>/updates")
@link_target(views.UpdatesInFeedLink, ["_system_id", "_feed_id"])
@requires_permissions(PermissionsLevel.ADMIN_READ)
def list_updates_in_feed(system_id, feed_id):
    """List recent feed updates."""
    return feedservice.list_updates_in_feed(system_id, feed_id)



"""
The permissions module is used to validate that certain HTTP requests have
permission to access a given HTTP resource.

Right now the Transiter permissions system is pretty basic. If works by
inspecting a HTTP header X-Transiter-AllowedMethods to determine what the
minimum allowable permission level is, and for a given resource verifies that
that level is met. The HTTP header is expected to be set by the reverse proxy
(for example, Nginx) that forwards requests to Transiter. This allows only
a subset of Transiter endpoints to be exposed outside the server.

When an administrator needs, they can bypass this system by SSH tunnelling
directly to the Transiter HTTP service.
"""
import enum

import flask
from decorator import decorator

from transiter import exceptions


class PermissionsLevel(enum.Enum):
    """
    The allowable permissions levels.

    Transiter permissions level are hierarchical: access to a given level
    permits access to all lower level.
    """

    USER_READ = 0
    ADMIN_READ = 1
    ALL = 2


def ensure(minimum_level):
    """
    Ensure the current HTTP request has sufficient permissions to access a
    resource.

    :param minimum_level: the minimum permissions level required to access the
                          resource
    :type minimum_level: PermissionsLevel
    :raises AccessDenied: if the permissions level is not met
    """
    request_level_str = flask.request.headers.get(
        "X-Transiter-PermissionsLevel", PermissionsLevel.ALL.name
    )
    try:
        request_level = PermissionsLevel[request_level_str.upper()]
    except KeyError:
        raise exceptions.InvalidPermissionsLevelInRequest(
            request_level_str.upper(),
            [permissions_level.name for permissions_level in PermissionsLevel],
        )
    if request_level.value < minimum_level.value:
        raise exceptions.AccessDenied(request_level.name, minimum_level.name)


def requires_permissions(minimum_level):
    """
    Decorator factory used to enforce permissions using decorators.

    :param minimum_level: the minimum level required for this endpoint
    :return: a decorator for the endpoint
    """

    @decorator
    def decorator_(func, *args, **kwargs):
        ensure(minimum_level)
        return func(*args, **kwargs)

    return decorator_



import enum

from sqlalchemy import Column, Integer, String, Enum, Boolean
from sqlalchemy.orm import relationship

from .base import Base


class System(Base):
    __tablename__ = "system"

    pk = Column(Integer, primary_key=True)
    id = Column(String, unique=True, index=True)

    class SystemStatus(enum.Enum):
        SCHEDULED = 1
        INSTALLING = 2
        ACTIVE = 3
        INSTALL_FAILED = 4
        DELETING = 5

    name = Column(String, nullable=True)
    status = Column(Enum(SystemStatus, native_enum=False), nullable=False)
    timezone = Column(String, nullable=True)
    auto_update_enabled = Column(Boolean, nullable=False, server_default="True")

    updates = relationship(
        "SystemUpdate", back_populates="system", cascade="all, delete-orphan"
    )
    routes = relationship(
        "Route", back_populates="system", cascade="all, delete-orphan"
    )
    stops = relationship("Stop", back_populates="system", cascade="all, delete-orphan")
    feeds = relationship("Feed", back_populates="system", cascade="all, delete-orphan")
    scheduled_services = relationship(
        "ScheduledService", back_populates="system", cascade="all, delete-orphan"
    )
    service_map_groups = relationship(
        "ServiceMapGroup", back_populates="system", cascade="all, delete-orphan"
    )



from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from .base import Base


class ScheduledTrip(Base):
    __tablename__ = "scheduled_trip"

    pk = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    route_pk = Column(Integer, ForeignKey("route.pk"), nullable=False)
    service_pk = Column(Integer, ForeignKey("scheduled_service.pk"), nullable=False)

    direction_id = Column(Boolean)

    route = relationship("Route", back_populates="scheduled_trips", cascade="none")
    service = relationship("ScheduledService", back_populates="trips", cascade="none")
    stop_times = relationship(
        "ScheduledTripStopTime", back_populates="trip", cascade="all, delete-orphan"
    )



"""
The GTFS Realtime Util contains the logic for reading feeds of this format.

The official reference is here: https://gtfs.org/reference/realtime/v2/
"""
import datetime

import pytz
from google.transit import gtfs_realtime_pb2

from transiter import parse


def create_parser(gtfs_realtime_pb2_module=None, post_pb2_parsing_function=None):
    """
    Create a GTFS Realtime parser.

    The returned parser performs the following steps in order:

    (1) Read the GTFS Realtime binary data into a JSON structure using either
        the default GTFS Realtime protobuf modules or a custom module.
        The custom is for when a GTFS Realtime extension is being used.

    (2) Optionally apply a post-pb2-parsing function to the JSON structure.
        This is generally use for merging GTFS Realtime extension data into
        the main JSON structure. Clean-up of the data within the feed is
        easier with a trip cleaner in (4).

    (3) Convert the JSON structure into various Transiter Python objects.
        These objects are SQL Alchemy models that correspond to database tables.

    :param gtfs_realtime_pb2_module: optional, the protobuffers module to use
        to read the feed. Defaults to the official module on PyPI.
    :param post_pb2_parsing_function: the function described in (2). It takes
        a single argument, which is the JSON data structure. It returns nothing;
        rather, it alterts the JSON structure in place.
    :return: the parser
    """
    # Additional arguments are accepted for forwards compatibility
    # noinspection PyUnusedLocal
    def parser(binary_content, *args, **kwargs):
        gtfs_data = read_gtfs_realtime(binary_content, gtfs_realtime_pb2_module)
        if post_pb2_parsing_function is not None:
            post_pb2_parsing_function(gtfs_data)
        (feed_time, trips) = transform_to_transiter_structure(
            gtfs_data, "America/New_York"
        )
        return trips

    return parser


built_in_parser = create_parser()


def read_gtfs_realtime(content, gtfs_realtime_pb2_module=None):
    """
    Convert a binary GTFS Realtime feed to a JSON-like data structure

    :param content: GTFS realtime binary data
    :param gtfs_realtime_pb2_module: GTFS realtime module
    :return: the data in a JSON-like data structure
    """
    if gtfs_realtime_pb2_module is None:
        gtfs_realtime_pb2_module = gtfs_realtime_pb2

    gtfs_feed = gtfs_realtime_pb2_module.FeedMessage()
    gtfs_feed.ParseFromString(content)
    return _read_protobuf_message(gtfs_feed)


def _read_protobuf_message(message):
    """
    Convert a protobuf message into a JSON-like structure.

    This is not an exhaustive converter but mean to be sufficient for GTFS
    realtime

    :param message: a google.protobuf.message.Message object
    :return: a dictionary of {key: value} for fields in the Message. The
    function is recursive so if value is also a protobuf then that will be
    expanded
    """
    d = {}
    for (descriptor, value) in message.ListFields():
        # Iterate through the possible protobuf field types to decide how to
        # parse this value
        if descriptor.type == descriptor.TYPE_MESSAGE:
            parsing_function = _read_protobuf_message
        elif descriptor.type == descriptor.TYPE_ENUM:

            def parsing_function(value):
                return descriptor.enum_type.values_by_number[value].name

        else:

            def parsing_function(value):
                return value

        # Then parse it
        if descriptor.label == descriptor.LABEL_REPEATED:
            parsed_value = [parsing_function(v) for v in value]
        else:
            parsed_value = parsing_function(value)

        d[descriptor.name] = parsed_value
    return d


if __name__ == "__main__":
    with open("path.gtfsrt", "rb") as f:
        result = read_gtfs_realtime(f.read())
    import json

    print(json.dumps(result, indent=2))


def transform_to_transiter_structure(data, timezone_str=None):
    """
    Transform GTFS Realtime data in JSON like format to Transiter models

    :param data: the data
    :param timezone_str: a string decribing the timezone
    :return: feed_time, trips, route_ids in the feed
    """
    transformer = _GtfsRealtimeToTransiterTransformer(data, timezone_str)
    return transformer.transform()


ACTIVE_PERIOD = "active_period"
AGENCY_ID = "agency_id"
ALERT = "alert"
CAUSE = "cause"
DESCRIPTION_TEXT = "description_text"
EFFECT = "effect"
END = "end"
ENTITY = "entity"
HEADER_TEXT = "header_text"
ID = "id"
INFORMED_ENTITY = "informed_entity"
ROUTE_ID = "route_id"
START = "start"
TRANSLATION = "translation"
TEXT = "text"
URL = "url"


class _GtfsRealtimeToTransiterTransformer:
    def __init__(self, raw_data, timezone_str=None):
        self._raw_data = raw_data
        self._trip_id_to_raw_entities = {}
        self._trip_id_to_transformed_entity = {}
        self._trip_id_to_trip_model = {}
        self._transformed_metadata = {}
        self._feed_time = None
        self._timestamp_to_datetime_cache = {}
        if timezone_str is None:
            self._timezone = None
        else:
            self._timezone = pytz.timezone(timezone_str)

    def transform(self):
        self._transform_feed_metadata()
        self._group_trip_entities()
        self._transform_trip_base_data()
        self._transform_trip_stop_events()
        self._update_stop_event_indices()
        return (
            self._feed_time,
            list(self._trip_id_to_trip_model.values())
            + self.build_alerts(self._raw_data),
        )

    def build_alerts(self, raw_data):
        def get_text(alert_data_, key):
            texts = alert_data_.get(key, {}).get(TRANSLATION, [])
            if len(texts) == 0:
                return None
            return texts[0][TEXT]

        def get_enum_value(enum, key, default):
            try:
                return enum[key]
            except KeyError:
                return default

        alerts = []
        for entity in raw_data.get(ENTITY, []):
            alert_id = entity.get(ID)
            alert_data = entity.get(ALERT)
            if alert_id is None or alert_data is None:
                continue
            informed_entities = alert_data.get(INFORMED_ENTITY, {})
            alert = parse.Alert(
                id=alert_id,
                cause=get_enum_value(
                    parse.Alert.Cause,
                    alert_data.get(CAUSE),
                    parse.Alert.Cause.UNKNOWN_CAUSE,
                ),
                effect=get_enum_value(
                    parse.Alert.Effect,
                    alert_data.get(EFFECT),
                    parse.Alert.Effect.UNKNOWN_EFFECT,
                ),
                header=get_text(alert_data, HEADER_TEXT),
                description=get_text(alert_data, DESCRIPTION_TEXT),
                url=get_text(alert_data, URL),
                start_time=self._timestamp_to_datetime(
                    alert_data.get(ACTIVE_PERIOD, {}).get(START)
                ),
                end_time=self._timestamp_to_datetime(
                    alert_data.get(ACTIVE_PERIOD, {}).get(START)
                ),
                agency_ids=[
                    informed_entity[AGENCY_ID]
                    for informed_entity in informed_entities
                    if AGENCY_ID in informed_entity
                ],
                route_ids=[
                    informed_entity[ROUTE_ID]
                    for informed_entity in informed_entities
                    if ROUTE_ID in informed_entity
                ],
            )
            alerts.append(alert)
        return alerts

    def _transform_feed_metadata(self):
        self._feed_time = self._timestamp_to_datetime(
            self._raw_data["header"]["timestamp"]
        )
        self._transformed_metadata = {"timestamp": self._feed_time}

    def _group_trip_entities(self):
        def attach_entity(entity_key, entity):
            trip_descriptor = entity.get("trip", {})
            trip_id = trip_descriptor.get("trip_id", None)
            if trip_id is None:
                return
            self._trip_id_to_raw_entities.setdefault(trip_id, {})
            self._trip_id_to_raw_entities[trip_id]["trip"] = trip_descriptor
            self._trip_id_to_raw_entities[trip_id][entity_key] = entity

        for main_entity in self._raw_data.get("entity", []):
            if "trip_update" in main_entity:
                attach_entity("trip_update", main_entity["trip_update"])
            if "vehicle" in main_entity:
                attach_entity("vehicle", main_entity["vehicle"])
            # if "alert" in main_entity:
            #    self._alerts_raw_data.append(main_entity["alert"])

    def _transform_trip_base_data(self):
        for trip_id, entity in self._trip_id_to_raw_entities.items():
            trip_data = entity.get("trip", {})
            vehicle_data = entity.get("vehicle", {})

            start_date_str = trip_data.get("start_date", None)
            if start_date_str is not None:
                start_dt = datetime.datetime(
                    year=int(start_date_str[0:4]),
                    month=int(start_date_str[4:6]),
                    day=int(start_date_str[6:8]),
                )
                trip_start_time = self._localize_datetime(start_dt, naive=True)
            else:
                trip_start_time = None

            raw_current_status = vehicle_data.get("current_status", None)
            if raw_current_status is not None:
                current_status = parse.Trip.Status[raw_current_status]
            else:
                current_status = None

            trip = parse.Trip(
                id=trip_id,
                route_id=trip_data.get("route_id", None),
                direction_id=trip_data.get("direction_id", None),
                start_time=trip_start_time,
                train_id=entity.get("trip_update", {})
                .get("vehicle", {})
                .get("id", None),
                updated_at=self._timestamp_to_datetime(
                    vehicle_data.get("timestamp", None)
                ),
                current_status=current_status,
                current_stop_sequence=vehicle_data.get("current_stop_sequence", 0),
                current_stop_id=vehicle_data.get("stop_id", None),
            )
            self._trip_id_to_trip_model[trip_id] = trip

    def _transform_trip_stop_events(self):
        for trip_id, trip in self._trip_id_to_trip_model.items():
            entity = self._trip_id_to_raw_entities[trip_id]

            trip_update = entity.get("trip_update", {})
            stop_time_updates = []

            for stop_time_update_data in trip_update.get("stop_time_update", []):
                stop_time_update = parse.TripStopTime(
                    stop_id=stop_time_update_data["stop_id"],
                    arrival_time=self._timestamp_to_datetime(
                        stop_time_update_data.get("arrival", {}).get("time", None)
                    ),
                    arrival_delay=stop_time_update_data.get("arrival", {}).get("delay"),
                    arrival_uncertainty=stop_time_update_data.get("arrival", {}).get(
                        "uncertainty"
                    ),
                    departure_time=self._timestamp_to_datetime(
                        stop_time_update_data.get("departure", {}).get("time", None)
                    ),
                    departure_delay=stop_time_update_data.get("departure", {}).get(
                        "delay"
                    ),
                    departure_uncertainty=stop_time_update_data.get(
                        "departure", {}
                    ).get("uncertainty"),
                    track=stop_time_update_data.get("track", None),
                )
                stop_time_updates.append(stop_time_update)
            trip.stop_times = stop_time_updates

    def _update_stop_event_indices(self):
        for trip_id, trip in self._trip_id_to_trip_model.items():
            index = trip.current_stop_sequence
            for stop_time_update in trip.stop_times:
                stop_time_update.stop_sequence = index
                index += 1

    def _timestamp_to_datetime(self, timestamp):
        if timestamp is None or timestamp == 0:
            return None
        if timestamp not in self._timestamp_to_datetime_cache:
            utc_dt_naive = datetime.datetime.utcfromtimestamp(timestamp)
            utc_dt = pytz.UTC.localize(utc_dt_naive)
            self._timestamp_to_datetime_cache[timestamp] = self._localize_datetime(
                utc_dt
            )
        return self._timestamp_to_datetime_cache[timestamp]

    def _localize_datetime(self, dt, naive=False):
        if self._timezone is None:
            return dt
        if naive:
            return self._timezone.localize(dt)
        else:
            return dt.astimezone(self._timezone)



import datetime
import enum
import typing
from dataclasses import dataclass, field


@dataclass
class Route:
    class Type(enum.Enum):
        LIGHT_RAIL = 0
        SUBWAY = 1
        RAIL = 2
        BUS = 3
        FERRY = 4
        CABLE_CAR = 5
        GONDOLA = 6
        FUNICULAR = 7

    id: str
    type: Type
    short_name: str = None
    long_name: str = None
    description: str = None
    color: str = None
    text_color: str = None
    url: str = None
    sort_order: int = None


@dataclass
class Stop:
    id: str
    name: str
    longitude: float
    latitude: float
    is_station: bool
    url: str = None
    parent_stop: typing.Optional["Stop"] = None


@dataclass
class ScheduledService:
    id: str
    monday: bool
    tuesday: bool
    wednesday: bool
    thursday: bool
    friday: bool
    saturday: bool
    sunday: bool
    # start_date: datetime.date = None
    # end_date: datetime.date = None
    trips: typing.List["ScheduledTrip"] = field(default_factory=list)
    # added_dates: typing.List[datetime.date] = field(default_factory=list)
    # removed_dates: typing.List[datetime.date] = field(default_factory=list)

    @classmethod
    def create_empty(cls, id_) -> "ScheduledService":
        return cls(
            id=id_,
            monday=False,
            tuesday=False,
            wednesday=False,
            thursday=False,
            friday=False,
            saturday=False,
            sunday=False,
        )


@dataclass
class ScheduledTrip:
    id: str
    route_id: str
    direction_id: bool
    stop_times: typing.List["ScheduledTripStopTime"] = field(default_factory=list)


@dataclass
class ScheduledTripStopTime:
    stop_id: str
    arrival_time: typing.Optional[datetime.time]
    departure_time: typing.Optional[datetime.time]
    stop_sequence: int


@dataclass
class Trip:
    class Status(enum.Enum):
        SCHEDULED = 1
        INCOMING_AT = 2
        STOPPED_AT = 3
        IN_TRANSIT_TO = 4

    id: str
    route_id: typing.Optional[str]
    direction_id: typing.Optional[bool]
    start_time: datetime.datetime = None
    stop_times: typing.List["TripStopTime"] = field(default_factory=list)

    # The following 5 fields correspond to VehiclePosition in the GTFS Realtime spec
    # and will be eventually moved to a new Vehicle type.
    train_id: str = None  # ?
    current_stop_sequence: int = None
    current_status: Status = None
    current_stop_id: str = None
    updated_at: datetime.datetime = None


@dataclass
class TripStopTime:
    stop_id: str
    arrival_time: datetime.datetime = None
    arrival_delay: int = None
    arrival_uncertainty: int = None
    departure_time: datetime.datetime = None
    departure_delay: int = None
    departure_uncertainty: int = None
    track: str = None  # Transiter-only non-GTFS field
    stop_sequence: int = None
    future: bool = True


@dataclass
class Alert:
    class Cause(enum.Enum):
        UNKNOWN_CAUSE = 1
        OTHER_CAUSE = 2
        TECHNICAL_PROBLEM = 3
        STRIKE = 4
        DEMONSTRATION = 5
        ACCIDENT = 6
        HOLIDAY = 7
        WEATHER = 8
        MAINTENANCE = 9
        CONSTRUCTION = 10
        POLICE_ACTIVITY = 11
        MEDICAL_EMERGENCY = 12

    class Effect(enum.Enum):
        NO_SERVICE = 0
        REDUCED_SERVICE = 1
        SIGNIFICANT_DELAYS = 2
        DETOUR = 3
        ADDITIONAL_SERVICE = 4
        MODIFIED_SERVICE = 5
        OTHER_EFFECT = 6
        UNKNOWN_EFFECT = 7
        STOP_MOVED = 8

    id: str
    header: str
    description: str
    start_time: datetime.datetime = None
    end_time: datetime.datetime = None
    cause: Cause = None
    effect: Effect = None
    url: str = None
    priority: int = None  # Transiter-only non-GTFS field

    agency_ids: typing.List[str] = field(default_factory=list)
    route_ids: typing.List[str] = field(default_factory=list)
    route_types: typing.List[Route.Type] = field(default_factory=list)
    trip_ids: typing.List[str] = field(default_factory=list)
    stop_ids: typing.List[str] = field(default_factory=list)


@dataclass
class DirectionRule:
    name: str
    id: str = None
    priority: int = None
    stop_id: str = None
    direction_id: bool = None
    track: str = None



import logging
import os

import flask
from werkzeug import exceptions as werkzeug_exceptions

from transiter import config, exceptions
from transiter.http import httpviews
from transiter.http.httpmanager import link_target, HttpStatus

docs_endpoints = flask.Blueprint(__name__, __name__)

logger = logging.getLogger(__name__)

_MISCONFIGURED_DOCUMENTATION_MESSAGE = """
The documentation appears to be misconfigured.

Please ensure that you have performed the following steps:

1. Generated the HTML documentation by executing `mkdocs build` in the
   Transiter documentation directory (./docs in the Git repo).

2. Set TRANSITER_DOCUMENTATION to point to the 'site' directory created by
   step 1. This environment variable can either be absolute, or relative
   to the directory of the Transiter WSGI app, which is:

       {}
"""

_FILE_NOT_FOUND_MESSAGE = """
Could not find file 'index.html' in the documentation root:

    TRANSITER_DOCUMENTATION_ROOT={}

The full absolute file path is:

    {}

Note that if TRANSITER_DOCUMENTATION_ROOT is a relative path, the absolute path is
calculated relative to the Transiter WSGI app's directory, which is:

    {}
"""


# NOTE: the built documentation refers to other HTML pages and assets using relative
# links. We need to redirect onto the URL with the slash in order for these to work.
@docs_endpoints.route("/", strict_slashes=True)
@docs_endpoints.route("/<path:path>", strict_slashes=True)
@link_target(httpviews.InternalDocumentationLink)
def docs(path="index.html", retry_with_index_dot_html=True, perform_validation=True):
    if not config.DOCUMENTATION_ENABLED:
        flask.abort(HttpStatus.NOT_FOUND)
    if perform_validation and not _documentation_root_is_valid():
        logger.error(
            _MISCONFIGURED_DOCUMENTATION_MESSAGE.format(
                flask.helpers.current_app.root_path
            )
        )
        raise exceptions.InternalDocumentationMisconfigured
    try:
        return flask.send_from_directory(config.DOCUMENTATION_ROOT, path)
    except werkzeug_exceptions.HTTPException:
        if path == "404.html":
            logger.error("404.html page is missing from documentation root!")
            raise exceptions.InternalDocumentationMisconfigured
        if not retry_with_index_dot_html:
            return docs(
                "404.html", retry_with_index_dot_html=False, perform_validation=False
            )
        if not path.endswith("/"):
            path += "/"
        path += "index.html"
        return docs(path, retry_with_index_dot_html=False, perform_validation=False)


# The verification string is included in index.html using the footer configuration
# in docs/mkdocs.yml.
_VERIFICATION_STRING = (
    "6808e24437044a523cf793e1f93f6924-"
    "90388b1a0c91760551522c59a4088557-"
    "7a55dcd8cb6f2093d5ab9a9b7997d9cc"
)


def _documentation_root_is_valid():
    filename = os.path.join(config.DOCUMENTATION_ROOT, "index.html")
    if not os.path.isabs(filename):
        filename = os.path.join(flask.helpers.current_app.root_path, filename)
    try:
        with open(filename) as f:
            return _VERIFICATION_STRING in f.read()
    except FileNotFoundError:
        logger.error(
            _FILE_NOT_FOUND_MESSAGE.format(
                config.DOCUMENTATION_ROOT, filename, flask.helpers.current_app.root_path
            )
        )
        return False



"""
The update manager contains the algorithm for executing feed updates.

This algorithm performs the following steps in order:

(1) Gets the parser for the Feed - either the built in parser or
    the custom parser. If this fails the update fails with explanation
    INVALID_PARSER.

(2) Downloads the data from the URL specified in the Feed. If this
    fails the update fails with explanation DOWNLOAD_ERROR. If the content
    of the download is empty, the update fails with explanation EMPTY_FEED.

(3) Calculate the hash of the downloaded data and compare it to the data
    used for the last successful FeedUpdate for this Feed. If they match,
    the update succeeds with explanation NOT_NEEDED.

(4) Perform the actual parsing using the Feed's parser. This "converts" the raw
    feed content into an iterator of UpdatableEntities. If any exception is
    raised by the parser then the update is failed with explanation
    PARSE_ERROR.

(5) Sync the results of the parser to the database using the sync module. If any
    exception is raised here, the update fails with explanation SYNC_ERROR.

(6) Otherwise, the update is deemed to be successful with explanation UPDATED.
"""
import datetime
import hashlib
import importlib
import json
import logging
import time
import traceback
import typing

import requests
from requests import RequestException

from transiter import models
from transiter.data import dbconnection
from transiter.data.dams import feeddam
from transiter.executor import celeryapp
from transiter.parse import parser
from transiter.services.update import sync
from . import gtfsrealtimeparser, gtfsstaticparser

logger = logging.getLogger(__name__)


def create_feed_update(system_id, feed_id) -> typing.Optional[int]:
    return _create_feed_update_helper(
        system_id, feed_id, update_type=models.FeedUpdate.Type.REGULAR
    )


def create_feed_flush(system_id, feed_id) -> typing.Optional[int]:
    return _create_feed_update_helper(
        system_id, feed_id, update_type=models.FeedUpdate.Type.FLUSH
    )


@dbconnection.unit_of_work
def _create_feed_update_helper(system_id, feed_id, update_type) -> typing.Optional[int]:
    feed = feeddam.get_in_system_by_id(system_id, feed_id)
    if feed is None:
        return None
    feed_update = models.FeedUpdate()
    feed_update.update_type = update_type
    feed_update.status = feed_update.Status.SCHEDULED
    feed_update.feed = feed
    dbconnection.get_session().flush()
    return feed_update.pk


@celeryapp.app.task
def execute_feed_update_async(feed_update_pk, content=None):
    return execute_feed_update(feed_update_pk, content)


def execute_feed_update(
    feed_update_pk, content=None
) -> typing.Tuple[models.FeedUpdate.Status, models.FeedUpdate.Result]:
    """
    Execute a feed update with logging and timing.

    For a description of what feed updates involve consult the module docs.
    """

    start_time = time.time()
    try:
        feed_update = _execute_feed_update_helper(feed_update_pk, content)
    except Exception as e:
        feed_update = models.FeedUpdate(
            pk=feed_update_pk,
            status=models.FeedUpdate.Status.FAILURE,
            result=models.FeedUpdate.Result.UNEXPECTED_ERROR,
            result_message=str(traceback.format_exc()),
        )
        logger.exception("Unexpected error", e)
    feed_update.total_duration = time.time() - start_time
    feed_update.completed_at = datetime.datetime.utcnow()
    with dbconnection.inline_unit_of_work() as session:
        session.merge(feed_update)
    logger.info(
        "{:7} / {:13} {:.2f} seconds / feed pk = {}.".format(
            feed_update.status.name,
            feed_update.result.name,
            feed_update.total_duration,
            feed_update.feed_pk,
        )
    )
    return feed_update.status, feed_update.result


class _InvalidParser(ValueError):
    """Exception raised when the Feed's parser is invalid."""

    pass


def _execute_feed_update_helper(feed_update_pk: int, content=None) -> models.FeedUpdate:
    """
    Execute a feed update.

    For a description of what feed updates involve consult the module docs.
    """
    with dbconnection.inline_unit_of_work() as session:
        feed_update = feeddam.get_update_by_pk(feed_update_pk)
        feed = feed_update.feed
        feed_update.status = feed_update.Status.IN_PROGRESS
        # We need to flush the session to persist the status change.
        session.flush()
        session.expunge(feed_update)
        session.expunge(feed)

    if feed_update.update_type == models.FeedUpdate.Type.FLUSH:
        return _sync_entities(feed_update, parser.CallableBasedParser(lambda: []))

    try:
        parser_object = _get_parser(feed.built_in_parser, feed.custom_parser)
    except _InvalidParser:
        return _update_status(
            feed_update,
            models.FeedUpdate.Status.FAILURE,
            models.FeedUpdate.Result.INVALID_PARSER,
            str(traceback.format_exc()),
        )

    if content is None:
        try:
            download_start_time = time.time()
            content = _get_content(feed.url, feed.headers)
        except RequestException as download_error:
            return _update_status(
                feed_update,
                models.FeedUpdate.Status.FAILURE,
                models.FeedUpdate.Result.DOWNLOAD_ERROR,
                str(download_error),
            )
        feed_update.download_duration = time.time() - download_start_time

    feed_update.content_length = len(content)
    if feed_update.content_length == 0:
        return _update_status(
            feed_update,
            models.FeedUpdate.Status.FAILURE,
            models.FeedUpdate.Result.EMPTY_FEED,
        )

    feed_update.content_hash = _calculate_content_hash(content)
    with dbconnection.inline_unit_of_work():
        previous_hash = feeddam.get_last_successful_update_hash(feed.pk)
    if previous_hash is not None and previous_hash == feed_update.content_hash:
        return _update_status(
            feed_update,
            models.FeedUpdate.Status.SUCCESS,
            models.FeedUpdate.Result.NOT_NEEDED,
        )

    # We catch any exception that can be thrown in the feed parser as, from our
    # perspective, the feed parser is foreign code.
    # noinspection PyBroadException
    try:
        parser_object.load_content(content)
    except Exception:
        return _update_status(
            feed_update,
            models.FeedUpdate.Status.FAILURE,
            models.FeedUpdate.Result.PARSE_ERROR,
            str(traceback.format_exc()),
        )

    return _sync_entities(feed_update, parser_object)


def _sync_entities(feed_update: models.FeedUpdate, parser_object):
    try:
        with dbconnection.inline_unit_of_work() as session:
            (
                feed_update.num_added_entities,
                feed_update.num_updated_entities,
                feed_update.num_deleted_entities,
            ) = sync.sync(feed_update.pk, parser_object)
            feed_update.num_parsed_entities = -1
            feed_update.status = models.FeedUpdate.Status.SUCCESS
            feed_update.result = models.FeedUpdate.Result.UPDATED
            session.merge(feed_update)
            return feed_update
    except Exception as e:
        logger.exception("Unexpected sync error", e)
        return _update_status(
            feed_update,
            models.FeedUpdate.Status.FAILURE,
            models.FeedUpdate.Result.SYNC_ERROR,
            str(traceback.format_exc()),
        )


_built_in_parser_to_function = {
    models.Feed.BuiltInParser.GTFS_STATIC: gtfsstaticparser.parse_gtfs_static,
    models.Feed.BuiltInParser.GTFS_REALTIME: gtfsrealtimeparser.built_in_parser,
}


def _update_status(feed_update: models.FeedUpdate, status, result, result_message=None):
    feed_update.status = status
    feed_update.result = result
    if result_message is not None:
        feed_update.result_message = result_message
    return feed_update


def _get_parser(built_in_parser, custom_parser) -> parser.TransiterParser:
    """
    Get the parser for a feed.
    """
    if built_in_parser is not None:
        return parser.cast_object_to_instantiated_transiter_parser(
            _built_in_parser_to_function[built_in_parser]
        )

    colon_char = custom_parser.find(":")
    if colon_char == -1:
        raise _InvalidParser(
            "Custom parser specifier must be of the form module:method"
        )
    module_str = custom_parser[:colon_char]
    method_str = custom_parser[(colon_char + 1) :]

    # The broad exception here is to capture any buggy code in the module being
    # imported.
    # noinspection PyBroadException
    try:
        module = _import_module(module_str)
    except ModuleNotFoundError:
        raise _InvalidParser("Unknown module '{}'".format(module_str))
    except Exception:
        raise _InvalidParser(f"Failed to import module {module_str}")

    try:
        module_attr = getattr(module, method_str)
    except AttributeError:
        raise _InvalidParser(
            "Module '{}' has no method '{}'.".format(module_str, method_str)
        )

    try:
        return parser.cast_object_to_instantiated_transiter_parser(module_attr)
    except ValueError:
        raise _InvalidParser(
            "Attribute '{}' of module '{}' is not a valid Transiter parser.".format(
                method_str, module_str
            )
        )


def _import_module(module_str, invalidate_caches=True):
    """
    Import a module.

    With invalidate caches True, if the import fails caches will be invalidated
    and the import attempted once more. Otherwise only one attempt is made.
    """
    try:
        return importlib.import_module(module_str)
    except ModuleNotFoundError:
        if invalidate_caches:
            importlib.invalidate_caches()
            return _import_module(module_str, invalidate_caches=False)
        else:
            raise


def _get_content(feed_url, feed_headers):
    """
    Download data from the feed's URL.
    """
    request = requests.get(feed_url, timeout=4, headers=json.loads(feed_headers))
    request.raise_for_status()
    return request.content


def _calculate_content_hash(content):
    """
    Calculate the MD5 hash of binary data.
    """
    m = hashlib.md5()
    m.update(content)
    return m.hexdigest()



"""
This module contains the graph data structures used in Transiter.

The data structures here are somewhat complex, and potentially over-engineered, but
are an attempt to address the following issues:

(1) Many graphs in Transiter are immutable. An ordered graph cannot have its nodes
    edges changed, for example, without potentially breaking the ordering property.
    The same applied for a tree and a path. Thus, significant effort is made to
    support immutable graphs.

(2) On the other hand, enforcing immutability throughout the code base can be
    inefficient as it means small operations involve copying the graph to a new,
    mutable graph before the operation and, potentially, copying the results to
    a new immutable graph after the operation.

(3) Given the existence of an immutable graph type, an immutable tree type is more
    efficiently implemented by composing the graph rather than inheriting from it.
    This way the internal graph data structures don't need to be copied as part of
    the initialization operation.

This has led to the following class layout:

- An abstract class AbstractGraph which describes the basic operations on a graph.
    This class contains nodes of type AbstractNode.

- A concrete Graph class which implements AbstractGraph and provides an immutable
    graph type. Its nodes are of type Graph.Node.

- A concrete MutableGraph class which extends Graph to contain mutable methods. It
    uses nodes of type MutableGraph.Node, which support mutation.

- An semi-abstract class ComposedGraph which enables creating graphs that are
    based on other graphs by composition. ComposedGraph implements AbstractGraph. A
    ComposedGraph contains an immutable backing graph, and all of the AbstractGraph
    methods are delegated to this backing graph.

- Concrete OrderedGraph, Tree, OrderedTree and Path which extend ComposedGraph
    and provide additional features -- for example, iterating over the nodes in an
    OrderedGraph returns the nodes as given by the order.
"""

from abc import abstractmethod, ABCMeta
from collections import defaultdict
from typing import Iterator, Set, FrozenSet, Iterable, Tuple, Optional, Dict, Collection


class AbstractNode(metaclass=ABCMeta):
    """
    Interface for Transiter's node types.
    """

    label = None

    @property
    @abstractmethod
    def in_nodes(self) -> Collection["AbstractNode"]:
        """
        Return all nodes A such that (A, this node) is an edge of the graph.
        """

    @property
    @abstractmethod
    def out_nodes(self) -> Collection["AbstractNode"]:
        """
        Return all nodes A such that (this node, A) is an edge of the graph.
        """


class AbstractGraph(metaclass=ABCMeta):
    """
    Interface for Transiter's graph types.
    """

    @abstractmethod
    def reverse(self) -> None:
        """
        Reverse the orientation of the graph.

        This switches the orientation of each edge in the graph from (node_1 -> node_2)
        to (node_2 -> node_1). Implementations below achieve this in O(1).
        """

    @abstractmethod
    def edge_label_tuples(self) -> Iterator[Tuple]:
        """
        Return all tuples (label_1, label_2) corresponding to edges of the graph.
        """

    @abstractmethod
    def get_node(self, label) -> Optional[AbstractNode]:
        """
        Return a node given its label, or None if no such node exists.
        """

    @abstractmethod
    def nodes(self) -> Iterator[AbstractNode]:
        """
        Return all nodes in the graph.
        """

    @abstractmethod
    def sources(self) -> Iterator[AbstractNode]:
        """
        Return all nodes that have no in-edges.
        """

    @abstractmethod
    def immutable(self) -> "Graph":
        """
        Return an immutable graph that is identical to this graph.
        """

    @abstractmethod
    def mutable(self) -> "MutableGraph":
        """
        Return a mutable graph that is identical to this graph.
        """

    @abstractmethod
    def __len__(self):
        """
        Return the number of nodes in the graph.
        """

    def __eq__(self, other):
        if not isinstance(other, AbstractGraph):
            print("This graph is not equal to a non-AbstractGraph object")
            return False
        if len(self) != len(other):
            print(
                f"This graph has a different number of nodes ({len(self)})"
                f" than the other graph ({len(other)})"
            )
            return False
        for other_node in other.nodes():
            other_label = other_node.label
            self_node = self.get_node(other_label)
            if self_node is None:
                print(
                    f"The other graph has a node with label '{other_label}'"
                    f" that this node does not have"
                )
                return False
            self_out_node_labels = set(node.label for node in self_node.out_nodes)
            other_out_node_labels = set(node.label for node in other_node.out_nodes)
            if self_out_node_labels != other_out_node_labels:
                print(
                    "The other graph has a different edge set than this graph\n"
                    f"This: edges from {other_label} to {self_out_node_labels}\n"
                    f"Other: edges from {other_label} to {other_out_node_labels}"
                )
                return False
        return True


class Graph(AbstractGraph):
    """
    Immutable graph implementation of an abstract graph.
    """

    class Orientation:
        """
        Data structure that records the orientation of a directed graph.

        This data structure is basically a wrapper around a boolean variable which
        records whether the graph has its initial orientation (Orientation.INITIAL)
        or has been reversed (Orientation.REVERSED).

        The structure mainly exists for memory efficiency purposes. Instead of using
        this structure, it would be possible for each Node in a graph to hold a
        reference to the graph it is in, and then store the orientation as a variable
        of the graph. However this introduces a circular reference which makes the
        graph less to garbage collect. Instead, the graph and every node hold a
        reference to an instance of this class.
        """

        INITIAL = False
        REVERSED = True

        def __init__(self):
            self._orientation = self.INITIAL

        def reverse(self):
            self._orientation = self._orientation is False

        def __eq__(self, other):
            return self._orientation is other

    class Node(AbstractNode):
        """
        Data structure representing a node in an immutable graph.

        WARNING: this data structure should not be initialized directly. It should only
        exist and be used within the context of a graph. If you use it outside that
        context it may spontaneously lose all of its edges.
        """

        def __init__(self, label, orientation: "Graph.Orientation"):
            self.label = label
            self._adjacent_nodes = {
                True: None,
                False: None,
            }  # type: Dict[bool, Optional[FrozenSet[Graph.Node]]]
            self._orientation = orientation

        @property
        def in_nodes(self) -> FrozenSet["Graph.Node"]:
            return self._get_nodes(self._orientation == Graph.Orientation.INITIAL)

        @property
        def out_nodes(self) -> FrozenSet["Graph.Node"]:
            return self._get_nodes(self._orientation == Graph.Orientation.REVERSED)

        @in_nodes.setter
        def in_nodes(self, nodes: Iterator["Graph.Node"]) -> None:
            """
            Set the in nodes of this node.

            Note that in line with the data structure being immutable this operation
            can only be performed once.
            """
            self._set_nodes(self._orientation == Graph.Orientation.INITIAL, nodes)

        @out_nodes.setter
        def out_nodes(self, nodes: Iterator["Graph.Node"]) -> None:
            """
            Set the out nodes of this node.

            Note that in line with the data structure being immutable this operation
            can only be performed once.
            """
            self._set_nodes(self._orientation == Graph.Orientation.REVERSED, nodes)

        def _get_nodes(self, key) -> FrozenSet["Graph.Node"]:
            if self._adjacent_nodes[key] is None:
                raise RuntimeError(
                    "Attempting to access the edges of an immutable "
                    "node before they have been set."
                )
            return self._adjacent_nodes[key]

        def _set_nodes(self, key, nodes):
            if self._adjacent_nodes[key] is not None:
                raise RuntimeError(
                    "Attempting to set the edges of an immutable node again."
                )
            self._adjacent_nodes[key] = frozenset(nodes)

        def _clear_edges(self):
            """
            Remove all of the edges of this node. This method is only used to enable
            safer garbage collection and should in general not be used as it breaks the
            immutable nature of the data structure.
            """
            self._adjacent_nodes = {True: None, False: None}

        def __eq__(self, other):
            return self.label == other.label

        def __hash__(self):
            return hash(self.label)

        def __repr__(self):
            return (
                f"Graph.Node(label={self.label}, "
                f"in_nodes={list(node.label for node in self.in_nodes)}, "
                f"out_nodes={list(node.label for node in self.out_nodes)})"
            )

    def __init__(self, graph: "AbstractGraph" = None):
        self._orientation = self.Orientation()
        self._label_to_node = {}
        if graph is None:
            return
        self._init_label_to_node(
            graph.edge_label_tuples(), [node.label for node in graph.nodes()]
        )

    def _init_label_to_node(self, edge_label_tuples, additional_labels=None):
        label_to_in_nodes = defaultdict(lambda: set())
        label_to_out_nodes = defaultdict(lambda: set())

        if additional_labels is not None:
            for label in additional_labels:
                self._create_node(label)

        for label_1, label_2 in edge_label_tuples:
            for label in (label_1, label_2):
                self._create_node(label)
            label_to_in_nodes[label_2].add(self._label_to_node[label_1])
            label_to_out_nodes[label_1].add(self._label_to_node[label_2])

        for label in self._label_to_node.keys():
            self._label_to_node[label].in_nodes = label_to_in_nodes[label]
            self._label_to_node[label].out_nodes = label_to_out_nodes[label]

    def _create_node(self, label) -> "Graph.Node":
        if label not in self._label_to_node:
            self._label_to_node[label] = self.Node(label, self._orientation)
        return self._label_to_node[label]

    # Implementation of abstract methods

    def reverse(self) -> None:
        self._orientation.reverse()

    def edge_label_tuples(self) -> Iterator[Tuple]:
        for in_node in self.nodes():
            for out_node in in_node.out_nodes:
                yield in_node.label, out_node.label

    def get_node(self, label) -> Optional["Graph.Node"]:
        return self._label_to_node.get(label)

    def nodes(self) -> Iterator["Graph.Node"]:
        yield from self._label_to_node.values()

    def sources(self) -> Iterator["Graph.Node"]:
        for node in self.nodes():
            if len(node.in_nodes) == 0:
                yield node

    def immutable(self) -> "Graph":
        return self

    def mutable(self) -> "MutableGraph":
        return MutableGraph(self)

    # Class methods for creating immutable graphs

    @classmethod
    def build_from_edge_label_tuples(cls, tuples, additional_labels=None) -> "Graph":
        """
        Build a graph from a iterator of edge label tuples.
        """
        graph = cls()
        graph._init_label_to_node(tuples, additional_labels)
        return graph

    @classmethod
    def build_from_nodes(cls, nodes: Iterable[Node]) -> "Graph":
        """
        Build a graph from nodes of another graph.
        """
        labels = set()
        edge_label_tuples = set()
        for node in nodes:
            labels.add(node.label)
            for out_node in node.out_nodes:
                edge_label_tuples.add((node.label, out_node.label))
        return cls.build_from_edge_label_tuples(edge_label_tuples, labels)

    @classmethod
    def union(cls, graphs: Iterable[AbstractGraph]) -> "Graph":
        """
        Build a graph as a union of other graphs.
        """
        edge_label_tuples = set()
        for component in graphs:
            edge_label_tuples.update(component.edge_label_tuples())
        return cls.build_from_edge_label_tuples(edge_label_tuples)

    def __del__(self):
        """
        Perform pre-deletion memory clean-up.

        Before deleting the graph, clear all of the circular references in the graph's
        nodes. This ensures that the graph and all its dependencies can be garbage
        collected using Python's synchronous (i.e., reference count based) garbage
        collection procedure, rather than relying on the asynchronous procedure. As
        well as being more memory efficient, this helps avoid memory leaks.
        """
        for node in self.nodes():
            # I know we're being bold, but keeping _clear_edges protected is worth
            # it to prevent others being even more bold.
            # noinspection PyProtectedMember
            node._clear_edges()

    def __contains__(self, item):
        if isinstance(item, AbstractNode):
            item = item.label
        return item in self._label_to_node

    def __len__(self):
        return len(self._label_to_node)


class MutableGraph(Graph):
    """
    Mutable graph implementation of an abstract graph.
    """

    class Node(Graph.Node):
        """
        Mutable node type.
        """

        def __init__(self, label, orientation: "Graph.Orientation"):
            self.label = label
            self._adjacent_nodes = {True: set(), False: set()}
            self._orientation = orientation

        def _get_nodes(self, key) -> Set["MutableGraph.Node"]:
            if self._adjacent_nodes[key] is None:
                return set()
            return self._adjacent_nodes[key]

        def _set_nodes(self, key, nodes):
            self._adjacent_nodes[key] = set(nodes)

    def edge_tuples(self):
        """
        Return all (A, B) node tuples where A -> B is an edge.
        """
        for in_node in self.nodes():
            for out_node in in_node.out_nodes:
                yield in_node, out_node

    def create_node(self, label):
        """
        Create a node with the given label.
        """
        return self._create_node(label)

    def create_edge_using_labels(self, label_1, label_2) -> None:
        """
        Create an edge between two nodes identified by their labels.

        Raises a ValueError if the graph does not contain nodes with the given labels.
        """
        for label in (label_1, label_2):
            if label not in self._label_to_node:
                raise ValueError(f"Graph has no node with label '{label}'")
        node_1 = self._label_to_node[label_1]
        node_2 = self._label_to_node[label_2]
        node_1.out_nodes.add(node_2)
        node_2.in_nodes.add(node_1)

    def delete_node(self, node) -> None:
        """
        Delete a node from the graph, along with all edges it is in.
        """
        for in_node_label in set(in_node.label for in_node in node.in_nodes):
            self.delete_edge_using_labels(in_node_label, node.label)
        for out_node_label in set(out_node.label for out_node in node.out_nodes):
            self.delete_edge_using_labels(node.label, out_node_label)
        del self._label_to_node[node.label]

    def delete_edge_using_labels(self, label_1, label_2) -> None:
        """
        Delete an edge from the graph using the labels of the adjacent nodes.
        """
        for label in (label_1, label_2):
            if label not in self._label_to_node:
                raise ValueError(f"Graph has no node with label '{label}'")
        node_1 = self._label_to_node[label_1]
        node_2 = self._label_to_node[label_2]
        if node_1 not in node_2.in_nodes:
            raise ValueError(f"Graph has no node edge ({label_1}, {label_2})")
        node_2.in_nodes.remove(node_1)
        node_1.out_nodes.remove(node_2)

    def immutable(self) -> "Graph":
        return Graph(self)

    def mutable(self) -> "MutableGraph":
        return self


def _delegating_method_factory(method_name):
    def delegating_method(self, *args, **kwargs):
        return getattr(self.backing_graph, method_name)(*args, **kwargs)

    return delegating_method


# Methods are added dynamically using the delegating method factory.
# noinspection PyAbstractClass
class ComposedGraph(AbstractGraph):
    """
    Objects of this type are graphs with additional metadata.

    The backing graph is stored via composition, as opposed to inheritance.
    """

    def __init__(self, graph: AbstractGraph):
        self._backing_graph = graph.immutable()

    @property
    def backing_graph(self):
        return self._backing_graph

    # AbstractGraph's __abstractmethod__ variable exists!
    # noinspection PyUnresolvedReferences
    for method_name in AbstractGraph.__abstractmethods__:
        vars()[method_name] = _delegating_method_factory(method_name)


# noinspection PyAbstractClass
class Tree(ComposedGraph):
    """
    Rooted tree type.
    """

    def __init__(self, graph: AbstractGraph, root: AbstractNode):
        """
        Initialize a new tree.

        Note that this is a good faith constructor. No attempt is made to verify that
        the graph is a tree, or that the root is in fact the root of the tree.

        :param graph: the graph to cast as a Tree
        :param root: the root of the tree
        """
        ComposedGraph.__init__(self, graph)
        self.root = self.get_node(root.label)

    def is_out_tree(self) -> bool:
        return not self.is_out_tree()

    def is_in_tree(self) -> bool:
        return len(self.root.in_nodes) == 1

    def __eq__(self, other):
        return AbstractGraph.__eq__(self, other) and self.root.label == other.root.label


# noinspection PyAbstractClass
class OrderedGraph(ComposedGraph):
    """
    Graph with a defined ordering of nodes.
    """

    def __init__(self, graph: AbstractGraph, ordering: Iterator[AbstractNode]):
        ComposedGraph.__init__(self, graph)
        self._ordering = [self._backing_graph.get_node(node.label) for node in ordering]

    def nodes(self):
        yield from self._ordering

    def reverse(self) -> None:
        self._backing_graph.reverse()
        self._ordering.reverse()


# noinspection PyAbstractClass
class Path(OrderedGraph):
    """
    A path graph data type.
    """

    def __init__(self, graph: AbstractGraph, head: AbstractNode):
        ordering = [head]
        while len(head.out_nodes) > 0:
            head = next(iter(head.out_nodes))
            ordering.append(head)
        OrderedGraph.__init__(self, graph, ordering)

    @classmethod
    def build_from_label_list(cls, label_list):
        if len(set(label_list)) != len(label_list):
            raise ValueError("Label list must contain unique elements")
        edge_tuples = set()
        for i in range(len(label_list) - 1):
            edge_tuples.add((label_list[i], label_list[i + 1]))
        graph = Graph.build_from_edge_label_tuples(edge_tuples, label_list)
        return cls(graph, graph.get_node(label_list[0]))


# noinspection PyAbstractClass
class OrderedTree(OrderedGraph, Tree):
    """
    A type representing a rooted tree with a defined ordering.
    """

    def __init__(self, tree: Tree, ordering: Iterator[AbstractNode]):
        OrderedGraph.__init__(self, tree, ordering)
        self.root = tree.root

    def reverse(self) -> None:
        self._backing_graph.reverse()
        self._ordering.reverse()


class Stack:
    def __init__(self, initial_elements: Iterable = None):
        if initial_elements is None:
            initial_elements = []
        self._store = list(initial_elements)

    def push(self, element):
        self._store.append(element)

    def pop(self):
        if len(self._store) == 0:
            raise LookupError("Stack has zero elements; cannot pop.")
        return self._store.pop()

    def __len__(self):
        return len(self._store)



import datetime
import typing

from transiter import parse


class TransiterParser:
    """
    Transiter's public parser API.

    This class is used to implement arbitrary parsers in Transiter. These can be used
    to read any kind of transit data into Transiter as long as the data can be pivoted
    to transiter.parse data types.

    To create a custom parser, subclass this class. You must implement the load_content
    method. Other than that, you implement methods based on which types your parser
    can return. For example, if your parser returns data of type transiter.parse.Stop,
    implement the get_stops method to return these entities.
    """

    def load_content(self, content: bytes) -> None:
        """
        Load content from a feed into the parser.

        Note that in many parser implementations this will simply involve:

            self._content = content

        However many parsers have more logic at this point.
        """
        raise NotImplementedError  # pragma: no cover

    def get_timestamp(self) -> typing.Optional[datetime.datetime]:
        pass  # pragma: no cover

    def get_routes(self) -> typing.Iterable[parse.Route]:
        pass  # pragma: no cover

    def get_stops(self) -> typing.Iterable[parse.Stop]:
        pass  # pragma: no cover

    def get_scheduled_services(self) -> typing.Iterable[parse.ScheduledService]:
        pass  # pragma: no cover

    def get_trips(self) -> typing.Iterable[parse.Trip]:
        pass  # pragma: no cover

    def get_alerts(self) -> typing.Iterable[parse.Alert]:
        pass  # pragma: no cover

    def get_direction_rules(self) -> typing.Iterable[parse.DirectionRule]:
        pass  # pragma: no cover

    # The rest of this class is not a part of the public API, and may be changed!
    # Developers are strongly discouraged from changing any data types or methods
    # below when implementing this class.

    __type_to_method__ = {
        parse.Route: get_routes,
        parse.Stop: get_stops,
        parse.ScheduledService: get_scheduled_services,
        parse.Trip: get_trips,
        parse.Alert: get_alerts,
        parse.DirectionRule: get_direction_rules,
    }

    @property
    def supported_types(self):
        """
        Return a set of transiter.parse types that are supported by this parser.
        """
        return set(
            type_
            for type_ in self.__type_to_method__.keys()
            if self._get_entities_getter(type_) is not None
        )

    def get_entities(self, entity_type):
        """
        Return an iterator of entities of a given type.

        This method ultimately calls the relevant getter method above; for example,
        with entity_type=parse.Route, this returns the result of get_routes above.

        If the method has not been implemented, then a NotImplementedError is raised.
        """
        method = self._get_entities_getter(entity_type)
        if method is None:
            raise NotImplementedError(
                "This parser does not support entities of type transiter.parse.{}.".format(
                    entity_type.__name__
                )
            )
        # Using `iter` here instead of `yield from` means the NotImplementedError above
        # is raised as soon as the method is called, rather than when the iteration
        # begins.
        return iter(method())

    def _get_entities_getter(self, entity_type):
        """
        Get the entities getter corresponding to a given type.

        This method returns None if the method has not been implemented.
        """
        abstract_method = self.__type_to_method__[entity_type]
        method = getattr(self, abstract_method.__name__)
        if method.__func__ is abstract_method:
            return None
        return method


class CallableBasedParser(TransiterParser):
    """
    Parser that uses the results of a Python callable to generate entities.

    This implements Transiter's simple parser API, whereby a parser can be a simple
    callable that returns entities of type transiter.parser.
    """

    def __init__(self, callable_: typing.Callable):
        self._callable = callable_
        self._type_to_entities = {type_: [] for type_ in self.__type_to_method__.keys()}

    def load_content(self, content: bytes) -> None:
        for entity in self._callable(content):
            type_ = type(entity)
            if type_ not in self._type_to_entities:
                raise TypeError("Unsupported parser type {}!".format(type_))
            self._type_to_entities[type_].append(entity)

    @property
    def supported_types(self):
        return set(self.__type_to_method__.keys())

    def get_entities(self, entity_type):
        return self._type_to_entities[entity_type]

    def __eq__(self, other):
        return self._callable is other._callable


def cast_object_to_instantiated_transiter_parser(object_):
    """
    Given an object, return an instantiated Transiter parser associated to that object.

    If the object is a TransiterParser type, the object itself is returned.
    Otherwise, if it is a callable, a CallableBasedParser using that callable is
    returned. Otherwise, a ValueError will be raised.
    """
    if isinstance(object_, type) and issubclass(object_, TransiterParser):
        return object_()
    if callable(object_):
        return CallableBasedParser(object_)
    raise ValueError("Cannot instantiate {} as a Transiter parser.".format(object_))



"""
This module contains the actual Flask app and is a such the 'root' of the HTTP
server. All other HTTP endpoints are linked to the app via blueprints in this
module.
"""
import datetime
import logging

import flask
import pytz
import werkzeug.exceptions as werkzeug_exceptions

from transiter import config, exceptions, __metadata__
from transiter.http import endpoints, httpviews
from transiter.http.httpmanager import (
    http_endpoint,
    HttpStatus,
    convert_exception_to_error_response,
)
from transiter.services import systemservice

app = flask.Flask("transiter")

app.register_blueprint(endpoints.docs_endpoints, url_prefix="/docs")
app.register_blueprint(endpoints.admin_endpoints, url_prefix="/admin")
app.register_blueprint(
    endpoints.feed_endpoints, url_prefix="/systems/<system_id>/feeds"
)
app.register_blueprint(
    endpoints.route_endpoints, url_prefix="/systems/<system_id>/routes"
)
app.register_blueprint(
    endpoints.stop_endpoints, url_prefix="/systems/<system_id>/stops"
)
app.register_blueprint(endpoints.system_endpoints, url_prefix="/systems")
app.register_blueprint(
    endpoints.trip_endpoints, url_prefix="/systems/<system_id>/routes/<route_id>/trips"
)
app.url_map.strict_slashes = False

logger = logging.getLogger("transiter")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
logger.addHandler(handler)
formatter = logging.Formatter("%(asctime)s WS %(levelname)-5s [%(module)s] %(message)s")
handler.setFormatter(formatter)


@app.errorhandler(Exception)
def transiter_error_handler(exception: Exception):
    """
    Error handler for exceptions.
    """
    if not isinstance(exception, exceptions.TransiterException):
        exception = exceptions.UnexpectedError(exception)
    return convert_exception_to_error_response(exception)


@app.errorhandler(HttpStatus.NOT_FOUND)
def page_not_found(__=None):
    """
    What to return if a user requests an endpoint that doesn't exist.
    """
    return transiter_error_handler(exceptions.PageNotFound(flask.request.path))


@app.errorhandler(HttpStatus.METHOD_NOT_ALLOWED)
def method_not_allowed(werkzeug_exception: werkzeug_exceptions.MethodNotAllowed):
    """
    What to return if a user requests an endpoint with a disallowed method.
    """
    return transiter_error_handler(
        exceptions.MethodNotAllowed(
            flask.request.method, flask.request.path, werkzeug_exception.valid_methods
        )
    )


@http_endpoint(app, "/")
def root():
    """HTTP/REST API entry point.

    Provides basic information about this Transiter instance and the Transit
    systems it contains.
    """
    if config.DOCUMENTATION_ENABLED:
        documentation_link = httpviews.InternalDocumentationLink()
    else:
        documentation_link = httpviews.ExternalDocumentationLink()
    response = {
        "transiter": {
            "version": __metadata__.__version__,
            "href": "https://github.com/jamespfennell/transiter",
            "docs": documentation_link,
            "build": _generate_build_response(),
        },
        "systems": httpviews.SystemsInstalled(count=len(systemservice.list_all())),
    }
    return response


def _generate_build_response():
    if __metadata__.__build_number__ is None:
        return None
    human_time = (
        datetime.datetime.fromtimestamp(
            __metadata__.__build_timestamp__, pytz.timezone("US/Eastern")
        ).isoformat()
        if __metadata__.__build_timestamp__ is not None
        else None
    )
    return {
        "number": __metadata__.__build_number__,
        "built_at": human_time,
        "built_at_timestamp": __metadata__.__build_timestamp__,
        "git_commit_hash": __metadata__.__git_commit_hash__,
        "href": __metadata__.__build_href__,
    }


def launch(force=False):
    """
    Launch the Flask app in debug mode.
    """
    logger.setLevel(logging.DEBUG)
    app.run(port=8000, debug=True)



import flask

from transiter.http.httpmanager import link_target, http_endpoint
from transiter.services import tripservice, views

trip_endpoints = flask.Blueprint(__name__, __name__)


@http_endpoint(trip_endpoints, "")
def list_all_in_route(system_id, route_id):
    """List all trips for a specific system"""
    return tripservice.list_all_in_route(system_id, route_id)


@http_endpoint(trip_endpoints, "/<trip_id>")
@link_target(views.Trip, ["_system_id", "_route_id", "id"])
def get_in_route_by_id(system_id, route_id, trip_id):
    """Retrieve a specific trip in a specific system."""
    return tripservice.get_in_route_by_id(system_id, route_id, trip_id)



"""
Feed update column clean up.

Revision ID: b521e2a9055a
Revises: be93cf805bb4
Create Date: 2020-03-19 22:17:03.347841

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b521e2a9055a"
down_revision = "be93cf805bb4"
branch_labels = None
depends_on = None


def upgrade():
    # (1) Remove indices
    op.drop_index("feed_updates_last_successful_idx", table_name="feed_update")
    op.drop_index("ix_feed_update_last_action_time", table_name="feed_update")

    # (2) Remove existing columns
    op.drop_column("feed_update", "failure_message")
    op.drop_column("feed_update", "execution_duration")
    op.drop_column("feed_update", "feed_time")
    op.drop_column("feed_update", "explanation")
    op.drop_column("feed_update", "raw_data_hash")

    # (3) Rename last_action_time to completed_at
    op.alter_column(
        "feed_update",
        "last_action_time",
        new_column_name="completed_at",
        server_default=None,
    )

    # (4) Add new columns
    op.add_column(
        "feed_update",
        sa.Column("content_created_at", sa.TIMESTAMP(timezone=True), nullable=True),
    )
    op.add_column("feed_update", sa.Column("content_hash", sa.String(), nullable=True))
    op.add_column(
        "feed_update", sa.Column("download_duration", sa.Float(), nullable=True)
    )
    op.add_column(
        "feed_update",
        sa.Column(
            "result",
            sa.Enum(
                "UPDATED",
                "NOT_NEEDED",
                "PARSE_ERROR",
                "DOWNLOAD_ERROR",
                "INVALID_PARSER",
                "EMPTY_FEED",
                "SYNC_ERROR",
                "UNEXPECTED_ERROR",
                name="result",
                native_enum=False,
            ),
            nullable=True,
        ),
    )
    op.add_column(
        "feed_update", sa.Column("num_parsed_entities", sa.Integer(), nullable=True)
    )
    op.add_column(
        "feed_update", sa.Column("num_added_entities", sa.Integer(), nullable=True)
    )
    op.add_column(
        "feed_update", sa.Column("num_updated_entities", sa.Integer(), nullable=True)
    )
    op.add_column(
        "feed_update", sa.Column("num_deleted_entities", sa.Integer(), nullable=True)
    )
    op.add_column(
        "feed_update", sa.Column("result_message", sa.String(), nullable=True)
    )
    op.add_column(
        "feed_update",
        sa.Column("scheduled_at", sa.TIMESTAMP(timezone=True), nullable=True),
    )
    op.add_column("feed_update", sa.Column("total_duration", sa.Float(), nullable=True))

    # (5) Add multi-column indices
    op.create_index(
        "feed_update_success_pk_completed_at_idx",
        "feed_update",
        ["feed_pk", "completed_at"],
        unique=False,
        postgresql_where=sa.text("status = 'SUCCESS'"),
    )
    op.create_index(
        "feed_update_feed_pk_feed_update_pk_idx",
        "feed_update",
        ["feed_pk", "pk"],
        unique=False,
    )


def downgrade():
    pass



from sqlalchemy import Index, Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.DirectionRule)
class DirectionRule(Base):
    __tablename__ = "direction_name_rule"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    stop_pk = Column(Integer, ForeignKey("stop.pk"))
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    priority = Column(Integer)
    direction_id = Column(Boolean)
    track = Column(String)
    name = Column(String)

    source = relationship("FeedUpdate", cascade="none")
    stop = relationship("Stop", back_populates="direction_rules", cascade="none")

    __table_args__ = (
        Index("direction_name_rule_stop_pk_priority_idx", stop_pk, priority),
    )

    @staticmethod
    def from_parsed_direction_rule(parsed_direction_rule: parse.DirectionRule):
        return DirectionRule(
            id=parsed_direction_rule.id,
            priority=parsed_direction_rule.priority,
            direction_id=parsed_direction_rule.direction_id,
            track=parsed_direction_rule.track,
            name=parsed_direction_rule.name,
        )



import enum

from sqlalchemy import (
    Column,
    UniqueConstraint,
    Enum,
    Integer,
    String,
    Boolean,
    ForeignKey,
)
from sqlalchemy.orm import relationship

from .base import Base


class Feed(Base):
    __tablename__ = "feed"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    system_pk = Column(Integer, ForeignKey("system.pk"))

    class BuiltInParser(enum.Enum):
        GTFS_STATIC = 1
        GTFS_REALTIME = 2

    built_in_parser = Column(Enum(BuiltInParser, native_enum=False))
    custom_parser = Column(String)
    url = Column(String)
    headers = Column(String)
    auto_update_enabled = Column(Boolean, nullable=False)
    auto_update_period = Column(Integer)
    required_for_install = Column(Boolean, nullable=False, default=False)

    system = relationship("System", back_populates="feeds")
    updates = relationship(
        "FeedUpdate", back_populates="feed", cascade="all, delete-orphan"
    )

    __table_args__ = (UniqueConstraint("system_pk", "id"),)



from setuptools import setup, find_packages

metadata = {}
with open("transiter/__metadata__.py") as f:
    exec(f.read(), metadata)
version = metadata["__version__"]

setup(
    name="transiter",
    version=version,
    author="James Fennell",
    author_email="jamespfennell@gmail.com",
    description="HTTP web service for transit data",
    url="https://github.com/jamespfennell/transiter",
    packages=find_packages(),
    license="MIT",
    entry_points={"console_scripts": ["transiterclt = transiter.clt:transiter_clt"]},
    install_requires=[
        "alembic==1.4.2",
        "apscheduler==3.6.3",
        "celery==4.4.2",
        "click==7.1.1",
        "decorator==4.4.2",
        "flask==1.1.2",
        "gunicorn==20.0.4",
        "gtfs-realtime-bindings==0.0.6",
        "inflection==0.4.0",
        "Jinja2==2.11.2",
        "psycopg2-binary==2.8.5",
        "pytimeparse==1.1.8",
        "pytz==2019.3",
        "requests==2.23.0",
        "sqlalchemy==1.3.16",
        "strictyaml==1.0.6",
    ],
)



import enum
from typing import Iterator

from .datastructures import AbstractNode, Stack


class DfsTraversalOrder(enum.Enum):
    PRE_ORDER = 0
    POST_ORDER = 1


def pre_order_dfs_traversal(
    starting_node: AbstractNode, ignore_directionality=False, sorting_key=None,
) -> Iterator[AbstractNode]:

    yield from dfs_traversal(
        DfsTraversalOrder.PRE_ORDER, starting_node, ignore_directionality, sorting_key
    )


def post_order_dfs_traversal(
    starting_node: AbstractNode, ignore_directionality=False, sorting_key=None,
) -> Iterator[AbstractNode]:
    yield from dfs_traversal(
        DfsTraversalOrder.POST_ORDER, starting_node, ignore_directionality, sorting_key
    )


def dfs_traversal(
    order: DfsTraversalOrder,
    starting_node: AbstractNode,
    ignore_directionality=False,
    sorting_key=None,
) -> Iterator[AbstractNode]:
    visited_nodes = set()
    stack = Stack()
    stack.push((starting_node, False))
    while len(stack) > 0:
        node, second_visit = stack.pop()
        visited_nodes.add(node)
        if order == DfsTraversalOrder.POST_ORDER:
            if second_visit:
                yield node
                continue
            stack.push((node, True))
        else:
            yield node
        next_nodes = list(node.out_nodes)
        if ignore_directionality:
            next_nodes.extend(node.in_nodes)
        if sorting_key is not None:
            # The next nodes are placed on the stack in the order of the list here.
            # Therefore we reverse the list in the sorting to ensure they come off the
            # stack in the real order we want.
            next_nodes.sort(key=sorting_key, reverse=True)
        for next_node in next_nodes:
            if next_node not in visited_nodes:
                stack.push((next_node, False))



import flask
import requests

from transiter import exceptions
from transiter.http import httpmanager, httpviews
from transiter.http.httpmanager import (
    http_endpoint,
    link_target,
    HttpMethod,
    HttpStatus,
)
from transiter.http.permissions import requires_permissions, PermissionsLevel
from transiter.services import systemservice, views

system_endpoints = flask.Blueprint(__name__, __name__)


@http_endpoint(system_endpoints, "")
@link_target(httpviews.SystemsInstalled)
def list_all():
    """List all systems"""
    return systemservice.list_all()


@http_endpoint(system_endpoints, "/<system_id>")
@link_target(views.System, ["id"])
def get_by_id(system_id):
    """Get data on a specific system."""
    return systemservice.get_by_id(system_id)


@http_endpoint(
    system_endpoints, "/<system_id>", method=HttpMethod.PUT,
)
@requires_permissions(PermissionsLevel.ALL)
def install(system_id):
    """Install a system."""
    form_key_to_value = flask.request.form.to_dict()
    config_file_url = form_key_to_value.pop("config_file", None)

    sync = httpmanager.is_sync_request()
    system_update_pk = systemservice.install(
        system_id=system_id,
        config_str=_get_config_file(
            config_file_url, flask.request.files.get("config_file")
        ),
        extra_settings=form_key_to_value,
        config_source_url=config_file_url,
        sync=sync,
    )

    # This means the system already exists and nothing was done.
    if sync:
        status = HttpStatus.CREATED
    else:
        status = HttpStatus.ACCEPTED
    return systemservice.get_update_by_id(system_update_pk), status


def _get_config_file(config_source_url, uploaded_config_file):
    if config_source_url is not None:
        try:
            response = requests.get(config_source_url)
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise exceptions.InvalidInput(
                "Could not download YAML config file from '{}'".format(
                    config_source_url
                )
            )
        return response.text
    elif uploaded_config_file is not None:
        return uploaded_config_file.read().decode("utf-8")
    else:
        raise exceptions.InvalidInput("YAML config file not provided!")


@http_endpoint(
    system_endpoints,
    "/<system_id>",
    method=HttpMethod.DELETE,
    returns_json_response=False,
)
@requires_permissions(PermissionsLevel.ALL)
def delete_by_id(system_id):
    """Uninstall a system."""
    systemservice.delete_by_id(
        system_id, error_if_not_exists=True, sync=httpmanager.is_sync_request()
    )
    return flask.Response(response="", status=HttpStatus.NO_CONTENT, content_type="")


@http_endpoint(
    system_endpoints, "/<system_id>/auto-update", method=HttpMethod.PUT,
)
@requires_permissions(PermissionsLevel.ALL)
def set_auto_update_enabled(system_id):
    form_key_to_value = flask.request.form.to_dict()
    enabled = form_key_to_value.get("enabled")
    if enabled is None:
        raise exceptions.InvalidInput("The form variable 'enabled' is required")
    enabled = enabled.lower()
    if enabled not in {"false", "true"}:
        raise exceptions.InvalidInput(
            "The form variable 'enabled' has to be 'true' or 'false', not '{}'".format(
                enabled
            )
        )
    systemservice.set_auto_update_enabled(
        system_id, form_key_to_value["enabled"].lower() == "true"
    )
    return "", HttpStatus.NO_CONTENT



import itertools
import typing

from transiter import models, parse
from transiter.data import dbconnection
from transiter.data.dams import scheduledam, genericqueries


def sync_trips(feed_update, parsed_services: typing.List[parse.ScheduledService]):
    num_entities_deleted = delete_trips_associated_to_feed(feed_update.feed.pk)
    if len(parsed_services) == 0:
        return num_entities_deleted > 0

    session = dbconnection.get_session()

    service_id_to_pk = genericqueries.get_id_to_pk_map_by_feed_pk(
        models.ScheduledService, feed_update.feed.pk
    )
    route_id_to_pk = genericqueries.get_id_to_pk_map(
        models.Route, feed_update.feed.system.pk
    )
    trip_mappings = []
    for parsed_service in parsed_services:
        for trip in parsed_service.trips:
            trip_mappings.append(
                {
                    "id": trip.id,
                    "route_pk": route_id_to_pk[trip.route_id],
                    "service_pk": service_id_to_pk[parsed_service.id],
                    "direction_id": trip.direction_id,
                }
            )
    session.bulk_insert_mappings(models.ScheduledTrip, trip_mappings)

    trip_id_to_pk = scheduledam.get_trip_id_to_pk_map_by_feed_pk(feed_update.feed.pk)
    stop_id_to_pk = genericqueries.get_id_to_pk_map(
        models.Stop, feed_update.feed.system.pk
    )
    # NOTE: SQL Alchemy's bulk_insert_mappings can take up a huge amount of memory if
    # executed on a large collection of mappings. If executed on the NYC Subway's
    # collection of stop times, it uses up to 750mb of memory. Chunking solves this
    # and actually seems to make the process faster.
    all_trips_iter = itertools.chain.from_iterable(
        parsed_service.trips for parsed_service in parsed_services
    )
    for chunk_of_trips in split(all_trips_iter, 100):
        stop_time_mappings = []
        for trip in chunk_of_trips:
            for stop_time in trip.stop_times:
                if stop_time.stop_id not in stop_id_to_pk:
                    continue
                stop_time_mappings.append(
                    {
                        "trip_pk": trip_id_to_pk[trip.id],
                        "stop_pk": stop_id_to_pk[stop_time.stop_id],
                        "arrival_time": stop_time.arrival_time,
                        "departure_time": stop_time.departure_time,
                        "stop_sequence": stop_time.stop_sequence,
                    }
                )
        session.bulk_insert_mappings(models.ScheduledTripStopTime, stop_time_mappings)

    return num_entities_deleted > 0 or len(trip_mappings) > 0


def split(container, size):
    chunk = []
    for index, element in enumerate(container):
        chunk.append(element)
        if (index + 1) % size == 0:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


def delete_trips_associated_to_feed(feed_pk):

    session = dbconnection.get_session()
    num_stop_times_deleted = (
        session.query(models.ScheduledTripStopTime)
        .filter(
            models.ScheduledTrip.pk == models.ScheduledTripStopTime.trip_pk,
            models.ScheduledService.pk == models.ScheduledTrip.service_pk,
            models.FeedUpdate.pk == models.ScheduledService.source_pk,
            models.FeedUpdate.feed_pk == feed_pk,
        )
        .delete(synchronize_session=False)
    )
    num_trips_deleted = (
        session.query(models.ScheduledTrip)
        .filter(
            models.ScheduledService.pk == models.ScheduledTrip.service_pk,
            models.FeedUpdate.pk == models.ScheduledService.source_pk,
            models.FeedUpdate.feed_pk == feed_pk,
        )
        .delete(synchronize_session=False)
    )
    # session.flush()
    return num_stop_times_deleted + num_trips_deleted



"""
Service map manager has responsibility for constructing and displaying service
maps
"""
import datetime
import json
import logging
import typing
from typing import List, Set, Tuple

from transiter import models
from transiter.data import dbconnection
from transiter.data.dams import scheduledam, servicemapdam, stopdam, tripdam
from transiter.services import views
from transiter.services.servicemap import graphutils, conditions

logger = logging.getLogger(__name__)


def build_stop_pk_to_service_maps_response(
    stop_pks,
) -> typing.Dict[str, List[views.ServiceMapWithRoutes]]:
    """
    Build the service maps response used in the stop service.
    """
    stop_pks = list(stop_pks)
    stop_pk_to_service_map_group_id_to_routes = servicemapdam.get_stop_pk_to_group_id_to_routes_map(
        stop_pks
    )
    stop_pk_to_service_maps_response = {}
    for stop_pk in stop_pks:
        group_id_to_routes = stop_pk_to_service_map_group_id_to_routes[stop_pk]
        stop_pk_to_service_maps_response[stop_pk] = [
            views.ServiceMapWithRoutes(
                group_id, list(map(views.Route.from_model, routes))
            )
            for group_id, routes in group_id_to_routes.items()
        ]
        continue
    return stop_pk_to_service_maps_response


def build_route_service_maps_response(
    route_pk,
) -> typing.List[views.ServiceMapWithStops]:
    response = []
    for group, service_map in servicemapdam.list_groups_and_maps_for_stops_in_route(
        route_pk
    ):
        if service_map is not None:
            stops = [
                views.Stop.from_model(entry.stop) for entry in service_map.vertices
            ]
        else:
            stops = []
        response.append(views.ServiceMapWithStops(group.id, stops))
    return response


def calculate_paths_hash(paths):
    string_edges = set()
    for path in paths:
        if len(path) <= 1:
            continue
        for k in range(len(path) - 1):
            string_edges.add(str(path[k]) + "," + str(path[k + 1]))
    return str(tuple(sorted(string_edges)))


def calculate_changed_route_pks_from_hashes(
    route_pk_to_previous_hash, route_pk_to_new_hash
):
    all_route_pks = set(route_pk_to_new_hash.keys()).union(
        route_pk_to_previous_hash.keys()
    )
    changed_route_pks = set()
    for route_pk in all_route_pks:
        previous_hash = route_pk_to_previous_hash.get(route_pk, None)
        new_hash = route_pk_to_new_hash.get(route_pk, None)
        if previous_hash is not None and new_hash is not None:
            if previous_hash == new_hash:
                continue
        changed_route_pks.add(route_pk)
    return changed_route_pks


def calculate_realtime_service_map_for_route(route):
    """
    Build the realtime service map for a route

    :param route: the route
    :return: nothing; the service map is persisted in the database
    """
    # First find the realtime service map group, if it exists.
    realtime_service_map = None
    for service_map_group in route.system.service_map_groups:
        if service_map_group.source != models.ServiceMapGroup.ServiceMapSource.REALTIME:
            continue
        realtime_service_map = service_map_group
        break
    if realtime_service_map is None:
        return

    old_service_map = None
    for service_map in list(realtime_service_map.maps):
        if service_map.route_pk == route.pk:
            old_service_map = service_map
            break

    # Now actually build the map.
    stop_pk_to_station_pk = stopdam.get_stop_pk_to_station_pk_map_in_system(
        route.system.id
    )
    trip_pk_to_path = tripdam.get_trip_pk_to_path_map(route.pk)
    paths = set()
    for trip in route.trips:
        path = trip_pk_to_path.get(trip.pk, [])
        if not trip.direction_id:
            path.reverse()
        paths.add(tuple(stop_pk_to_station_pk[stop_pk] for stop_pk in path))
    logger.info("Building realtime service map for route {}.".format(route.id))
    try:
        service_map = _build_service_map_from_paths(paths)
    except graphutils.topologicalsort.ImpossibleToTopologicallySortGraph:
        logger.info("Could not topologically sort:\n{}".format(json.dumps(list(paths))))
        return

    # Delete the old service map for this route.
    if old_service_map is not None:
        dbconnection.get_session().delete(old_service_map)
        dbconnection.get_session().flush()

    service_map.route = route
    service_map.group = realtime_service_map


def calculate_scheduled_service_maps_for_system(system):
    """
    Build the schedule service maps for a system

    :param system: the system
    :return: nothing; the service maps are persisted in the database
    """
    stop_pk_to_station_pk = stopdam.get_stop_pk_to_station_pk_map_in_system(system.id)
    trip_pk_to_stop_pks = scheduledam.get_scheduled_trip_pk_to_path_in_system(system.pk)
    route_pk_to_trips = {}

    for (
        trip,
        start_time,
        end_time,
    ) in scheduledam.list_scheduled_trips_with_times_in_system(system.pk):
        trip.start_time = start_time
        trip.end_time = end_time
        if not trip.direction_id:
            trip_pk_to_stop_pks.get(trip.pk, []).reverse()
        trip.path = tuple(
            stop_pk_to_station_pk[stop_pk]
            for stop_pk in trip_pk_to_stop_pks.get(trip.pk, [])
        )
        if trip.route_pk not in route_pk_to_trips:
            route_pk_to_trips[trip.route_pk] = []
        route_pk_to_trips[trip.route_pk].append(trip)

    for service_map_group in system.service_map_groups:
        if service_map_group.source != models.ServiceMapGroup.ServiceMapSource.SCHEDULE:
            continue
        # Delete the old maps, using SQL Alchemy's delete-orphan cascade
        service_map_group.maps = []
        dbconnection.get_session().flush()
        if service_map_group.conditions is not None:
            conditions = json.loads(service_map_group.conditions)
        else:
            conditions = None
        matcher = _ScheduledTripMatcher(conditions)

        for route_pk, trips in route_pk_to_trips.items():
            path_to_count = {}
            num_trips = 0
            for trip in trips:
                if matcher.match(trip):
                    path_to_count.setdefault(trip.path, 0)
                    path_to_count[trip.path] += 1
                    num_trips += 1

            final_paths = {
                path
                for path, count in path_to_count.items()
                if count >= num_trips * service_map_group.threshold
            }
            service_map = _build_service_map_from_paths(final_paths)
            service_map.route_pk = route_pk
            service_map.group = service_map_group


def _build_service_map_from_paths(paths):
    """
    Given a list of paths build the service map.
    """
    labels = []
    for graph in _build_sorted_graph_from_paths(paths):
        for node in graph.nodes():
            labels.append(node.label)
    return _convert_sorted_graph_to_service_pattern(labels)


def _convert_sorted_graph_to_service_pattern(sorted_labels):
    """
    Convert a sorted graph object to a service map object.
    """
    service_pattern = models.ServiceMap()
    for index, label in enumerate(sorted_labels):
        sp_vertex = models.ServiceMapVertex()
        sp_vertex.stop_pk = label
        sp_vertex.map = service_pattern
        sp_vertex.position = index
    return service_pattern


def _build_sorted_graph_from_paths(
    paths: Set[Tuple],
) -> List[graphutils.datastructures.AbstractGraph]:
    """
    Given a list of paths build the sorted graph.
    """
    if len(paths) == 0:
        return [graphutils.datastructures.Graph()]
    if len(paths) == 1:
        unique_path = next(iter(paths))
        if len(set(unique_path)) != len(unique_path):
            return [graphutils.datastructures.Graph()]
        return [graphutils.datastructures.Path.build_from_label_list(unique_path)]
    additional_nodes = set()
    edge_tuples = set()
    for path in paths:
        if len(path) == 1:
            additional_nodes.add(path[0])
            continue
        for i in range(len(path) - 1):
            edge_tuples.add((path[i], path[i + 1]))
    graph = graphutils.datastructures.MutableGraph.build_from_edge_label_tuples(
        edge_tuples, additional_nodes
    )
    return list(
        map(
            graphutils.topologicalsort.tgt_sort,
            graphutils.operations.split_into_connected_components(graph),
        )
    )


class _ScheduledTripMatcher:
    def __init__(self, raw_conds):
        if raw_conds is None:
            raw_conds = {}
        self._primary_conditions = self._convert_raw_conditions(raw_conds)

    def match(self, trip):
        for condition in self._primary_conditions:
            if not condition(trip):
                return False
        return True

    @classmethod
    def _convert_raw_conditions(cls, dictionary):
        conditions = []
        for key, value in dictionary.items():
            conditions.append(cls._convert_raw_condition(key, value))
        return conditions

    _logical_operators = None
    _key_to_function = None
    _key_to_extra_args = None

    @classmethod
    def _convert_raw_condition(cls, key, value):
        # Having these variables as class variables that are populated on the
        # first run is basically a form of caching
        if cls._logical_operators is None:
            cls._logical_operators = {
                conditions.ALL_OF,
                conditions.NONE_OF,
                conditions.ONE_OF,
            }
            cls._key_to_function = {
                conditions.ALL_OF: cls.all_of_factory,
                conditions.ENDS_EARLIER_THAN: cls.order_factory,
                conditions.ENDS_LATER_THAN: cls.order_factory,
                conditions.STARTS_EARLIER_THAN: cls.order_factory,
                conditions.NONE_OF: cls.none_of_factory,
                conditions.ONE_OF: cls.one_of_factory,
                conditions.STARTS_LATER_THAN: cls.order_factory,
                conditions.WEEKDAY: cls.weekday_factory,
                conditions.WEEKEND: cls.weekend_factory,
            }
            cls._key_to_extra_args = {
                conditions.ENDS_EARLIER_THAN: ("end_time", True),
                conditions.ENDS_LATER_THAN: ("end_time", False),
                conditions.STARTS_EARLIER_THAN: ("start_time", True),
                conditions.STARTS_LATER_THAN: ("start_time", False),
            }
        if key in cls._logical_operators:
            value = cls._convert_raw_conditions(value)
        try:
            func = cls._key_to_function[key]
        except KeyError:
            raise NotImplementedError
        extra_args = cls._key_to_extra_args.get(key, ())
        return func(value, *extra_args)

    @staticmethod
    def one_of_factory(conditions):
        def one_of(trip):
            for condition in conditions:
                if condition(trip):
                    return True
            return False

        return one_of

    @staticmethod
    def none_of_factory(conditions):
        def none_of(trip):
            for condition in conditions:
                if condition(trip):
                    return False
            return True

        return none_of

    @staticmethod
    def all_of_factory(conditions):
        def all_of(trip):
            for condition in conditions:
                if not condition(trip):
                    return False
            return True

        return all_of

    @staticmethod
    def order_factory(value, trip_attr, less_than=True):
        import math

        hour = int(math.floor(value))
        value = (value - hour) * 60
        minute = int(math.floor(value))
        value = (value - minute) * 60
        second = int(math.floor(value))
        t = datetime.time(hour=hour, minute=minute, second=second)

        def order(trip):
            attr = getattr(trip, trip_attr)
            if attr is None:
                return False
            return (attr < t) == less_than

        return order

    @staticmethod
    def weekday_factory(value):
        def weekday(trip):
            weekday_cond = (
                trip.service.monday
                or trip.service.tuesday
                or trip.service.wednesday
                or trip.service.thursday
                or trip.service.friday
            )
            weekend_cond = not (trip.service.saturday or trip.service.sunday)
            return (weekday_cond and weekend_cond) == value

        return weekday

    @classmethod
    def weekend_factory(cls, value):
        return cls.weekday_factory(not value)



import datetime
import enum

from sqlalchemy import (
    Column,
    Integer,
    String,
    Enum,
    TIMESTAMP,
    ForeignKey,
    Index,
    Float,
)
from sqlalchemy.orm import relationship

from .base import Base


class SystemUpdate(Base):
    __tablename__ = "system_update"

    pk = Column(Integer, primary_key=True)
    system_pk = Column(Integer, ForeignKey("system.pk"))

    class Status(enum.Enum):
        SCHEDULED = 1
        IN_PROGRESS = 2
        SUCCESS = 3
        FAILED = 4

    status = Column(Enum(Status, native_enum=False), nullable=False)
    status_message = Column(String)
    total_duration = Column(Float)
    scheduled_at = Column(TIMESTAMP(timezone=True), default=datetime.datetime.utcnow)
    completed_at = Column(TIMESTAMP(timezone=True))
    config = Column(String)
    config_template = Column(String)
    config_parameters = Column(String)
    config_source_url = Column(String)
    transiter_version = Column(String)

    system = relationship("System", back_populates="updates")

    __table_args__ = (
        Index("system_update_system_pk_system_update_pk_idx", system_pk, pk),
    )



import dataclasses

from transiter.services import views


@dataclasses.dataclass
class SystemsInstalled(views.View):
    count: int


@dataclasses.dataclass
class InternalDocumentationLink(views.View):
    pass


@dataclasses.dataclass
class ExternalDocumentationLink(views.View):
    href: str = "https://docs.transiter.io"
